
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>

#include <ngx_http_wasm_host.h>
#include <ngx_wasm_serial.h>
#include <ngx_wasm_log.h>
#include <ngx_wasm_filter.h>
#include <ngx_wasm_call_env.h>
#include <ngx_wasm_event.h>
#include <ngx_wasm_connection.h>

#include <wasi/ngx_wasi_http_request.h>
#include <wasi/ngx_wasi_http_module.h>


enum ngx_wasm_http_ops_tags_e {
    NGX_WASM_HTTP_TAG_REQUEST = 0x4854548, /* HTTH */
    NGX_WASM_HTTP_TAG_REQUEST_HEADERS,
    NGX_WASM_HTTP_TAG_REQUEST_BODY,
    NGX_WASM_HTTP_TAG_RESPONSE_HEADERS,
    NGX_WASM_HTTP_TAG_RESPONSE_BODY,
    NGX_WASM_HTTP_TAG_LOCATION_ENV_KV,
    NGX_WASM_HTTP_TAG_LOCATION_ENV_ARGV,
    NGX_WASM_HOST_CALL_ENV_API_TAG,
    NGX_WASM_HTTP_TAG_CONFIG
};


/*
 * wasm module that is loaded as HTTP handler: it can handle different
 * phases and contains init function that is called during worker init
 */
typedef struct {
    ngx_uint_t                        index;     /* wasm module index */
    ngx_uint_t                        own_index; /* in http wasm modules */
    ngx_str_t                         name;      /* for logging = module id */
    ngx_str_t                         phase_hooks[NGX_HTTP_LOG_PHASE + 1];
                                             /* phase handlers names, if any */
    ngx_str_t                         header_filter;
    ngx_str_t                         request_body_filter;
    ngx_str_t                         response_body_filter;
} ngx_http_wasm_module_t;


/*
 * WASM host used at configuration time - during worker init
 * provides interfaces to update angie configuration: i.e. install
 * phase handlers and similar
 */
typedef struct {
    ngx_wasm_host_t                   wh;       /* must be first */
    ngx_cycle_t                      *cycle;
    ngx_http_wasm_module_t           *hw_mod;
    int32_t                           conf_fd;
} ngx_http_wasm_host_conf_t;


/* instance of wasm http module */
typedef struct {
    ngx_wasm_instance_t              *wi;
    ngx_http_wasm_host_t             *host;
    ngx_http_wasm_module_t           *hw;

    int32_t                           host_fd;
    uint32_t                          wstate;

    ngx_wasm_call_env_t              *ce;

    ngx_event_t                       ev;
} ngx_http_wasm_module_state_t;


/* request context of this angie module */
typedef struct {
    /*
     * this HTTP angie module hosts multiple WASM HTTP modules;
     * each has its own request context; here is an array of them
     */
    ngx_http_wasm_module_state_t    **http_wasm_modules;

    /*
     * some events happen in context where only request is available;
     * thus we need to know who is responsible to handle them
     */
    ngx_http_wasm_module_state_t     *curr_module;
    ngx_uint_t                        curr_phase;

    ngx_uint_t                        next_phase_module;
    ngx_uint_t                        next_header_filter_module;
    ngx_uint_t                        next_request_body_filter_module;
    ngx_uint_t                        next_response_body_filter_module;

    /* state of request that is handled by wasm http modules */
    ngx_wasm_http_ctx_t               http_state;

    /* body filter */
    ngx_chain_t                      *free;
    ngx_chain_t                      *busy;
    ngx_chain_t                      *rb_out;
    ngx_chain_t                      *out;
    ngx_chain_t                      *in;
    ngx_uint_t                        rb_buffered; /* rb_buffered:1 */
} ngx_http_wasm_host_rctx_t;


typedef struct {
    ngx_wasm_func_conf_t              fc;
    ngx_array_t                       arg_vars; /* of ngx_complex_value_t */
    uint32_t                          wstate;
} ngx_http_wasm_host_var_ctx_t;


typedef struct {
    ngx_array_t                      *argv;       /* of cv */
    ngx_array_t                      *env_params; /* of env_param_t */
} ngx_http_wasm_env_ctx_t;


typedef struct {
    ngx_http_complex_value_t          key;
    ngx_http_complex_value_t          value;
} ngx_http_wasm_env_param_t;


/* request headers passed to wasm */
typedef struct {
    ngx_table_elt_t                  *te;
    ngx_list_part_t                  *part;
    ngx_uint_t                        off;
} ngx_wasm_http_headers_pos_t;


typedef struct {
    ngx_kv_array_t                    kva;
    ngx_wasm_http_headers_pos_t       pos;
} ngx_http_open_request_headers_t;


/* response headers read from wasm */
typedef struct {
    ngx_kv_array_t                    kva;
    ngx_http_request_t               *request;
} ngx_http_open_response_headers_t;


typedef struct {
    ngx_http_request_t               *r;
    ngx_http_request_body_t          *body;
    off_t                             offset;
    ngx_uint_t                        read_started;
    ngx_wasm_http_ctx_t              *host_ctx;
} ngx_http_open_request_body_t;


typedef struct {
    ngx_http_request_t               *r;

    ngx_chain_t                      *out_bufs;
    ngx_chain_t                      *free_bufs;
} ngx_http_open_response_body_t;


typedef struct {
    ngx_kv_array_t                   kva;
    ngx_str_array_t                  sa;
} ngx_http_wasm_env_state_t;


#if (NGX_API)

typedef struct {
    ngx_api_entry_t                  *root;
    ngx_connection_t                 *connection;
    ngx_str_t                         api_result;
    ngx_str_iter_t                    result_iter;
} ngx_wasm_api_ctx_t;

#endif

typedef struct {
    ngx_array_t                       modules; /* of ngx_http_wasm_module_t */
    ngx_array_t                       apis;
} ngx_http_wasm_host_main_conf_t;


typedef struct {
    ngx_http_wasm_env_ctx_t           env;

    /* all handler arrays contain index of http wasm module */
    ngx_array_t                       *header_filters;
    ngx_array_t                       *request_body_filters;
    ngx_array_t                       *response_body_filters;

    ngx_array_t                     *(*phase_handlers)[NGX_HTTP_LOG_PHASE + 1];

} ngx_http_wasm_host_loc_conf_t;


static ngx_http_wasm_host_t *ngx_http_wasm_host_create(ngx_wasm_instance_t *wi,
    ngx_http_request_t *r, const char *id);

/* request */
static int32_t ngx_wasm_http_request_open(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static ngx_inline int32_t ngx_http_get_u32(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len, uint32_t value);
static int32_t ngx_http_request_property_get(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
static uint32_t ngx_wasm_http_get_schema(ngx_http_request_t *r);
static int32_t ngx_http_request_property_var(ngx_wasm_host_t *wh,
    ngx_http_request_t *r, uint8_t *buf, uint32_t len);
static int32_t ngx_http_request_property_set(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
static ngx_int_t ngx_http_wasm_host_reject_request(ngx_http_request_t *r,
    ngx_str_t *reject_msg);
static int32_t ngx_wasm_http_request_close(ngx_wasm_handle_t *self);

/* request headers  */
static int32_t ngx_http_wasm_host_get_curr_header(uint64_t i, void *data,
    void *elt);
static int32_t ngx_http_wasm_host_next_header(void *data);
static void ngx_wasm_http_init_request_headers_encoder(ngx_kv_array_t *kva,
    ngx_list_part_t *part, ngx_wasm_http_headers_pos_t *pos);
static int32_t ngx_wasm_http_open_request_headers(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_read_request_headers(
    ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len);


/* request body */
static int32_t ngx_wasm_http_open_request_body(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_read_request_body(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len);
static ngx_int_t ngx_wasm_http_body_from_mem(ngx_http_request_body_t *body,
    off_t off, uint8_t *buf, size_t size, size_t *nread);
static ngx_int_t ngx_wasm_http_body_from_file(ngx_http_request_body_t *body,
    off_t off, uint8_t *buf, size_t size, size_t *nread, ngx_log_t *log);

/* response headers */
static int32_t ngx_http_wasm_init_response_header(ngx_arr_t *arr, uint64_t off,
    void *elt);
static int32_t ngx_wasm_http_open_response_headers(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_write_response_headers(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_http_set_response_headers(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);

/* response body */
static int32_t ngx_wasm_http_open_response_body(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_write_response_body(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_http_close_response_body(ngx_wasm_handle_t *self);

static char *ngx_http_wasm_env_argv(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static char *ngx_http_wasm_env_param(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);

static ngx_http_wasm_module_t *ngx_http_wasm_host_find_module(ngx_conf_t *cf,
    ngx_str_t *name);
static char *ngx_http_wasm_add_filter_handler(ngx_conf_t *cf,
    ngx_command_t *cmd, ngx_array_t **handlers);
static ngx_int_t ngx_http_wasm_add_phase_handler(ngx_conf_t *cf,
    ngx_uint_t phase, ngx_array_t **handlers, ngx_uint_t mod_index);

static char *ngx_http_wasm_handler(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static char *ngx_http_wasm_header_filter(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static char *ngx_http_wasm_request_body_filter(ngx_conf_t *cf,
    ngx_command_t *cmd, void *conf);
static char *ngx_http_wasm_response_body_filter(ngx_conf_t *cf,
    ngx_command_t *cmd, void *conf);


static ngx_array_t *ngx_http_wasm_host_create_call_env_kv(
    ngx_http_request_t *r);
static ngx_array_t *ngx_http_wasm_host_create_call_env_argv(
    ngx_http_request_t *r);
static int32_t ngx_wasm_http_open_location_env_kv(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_open_location_env_argv(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_http_read_location_env_kv(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_http_read_location_env_argv(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len);


#if (NGX_API)
static int32_t ngx_wasm_host_open_api(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static ngx_int_t ngx_wasm_host_api_get_result(ngx_wasm_api_ctx_t *ac,
    ngx_pool_t *pool, ngx_str_t *path, ngx_str_t *res);
static ngx_int_t ngx_wasm_host_api_decode_key(ngx_wasm_handle_t *self,
    uint8_t *buf, uint32_t len, ngx_str_t *res);
static int32_t ngx_wasm_host_api_get(ngx_wasm_handle_t *self, uint32_t propid,
    uint8_t *buf, uint32_t len);
#endif

static ngx_int_t ngx_http_wasm_host_add_ops(ngx_array_t *apis,
    ngx_wasm_handle_ops_t *ops);


static int32_t ngx_wasm_http_host_set_conf(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
static ngx_http_wasm_host_conf_t *ngx_http_wasm_conf_host_create(
    ngx_wasm_instance_t *wi, const char *id);
static ngx_int_t ngx_wasm_http_host_install_phase_hook(
    ngx_http_wasm_host_conf_t *host, ngx_uint_t phase, const char *phase_name,
    uint8_t *buf, uint32_t len);
static ngx_int_t ngx_wasm_http_host_install_request_body_filter(
    ngx_http_wasm_host_conf_t *host, uint8_t *buf, uint32_t len);
static ngx_int_t ngx_wasm_http_host_install_response_header_filter(
    ngx_http_wasm_host_conf_t *host, uint8_t *buf, uint32_t len);
static ngx_int_t ngx_wasm_http_host_install_response_body_filter(
    ngx_http_wasm_host_conf_t *host, uint8_t *buf, uint32_t len);

static ngx_int_t ngx_http_wasm_host_call_module(ngx_http_request_t *r,
    ngx_http_wasm_module_state_t *ctx, ngx_str_t *handler);
static ngx_http_wasm_host_rctx_t *ngx_http_wasm_host_get_ctx(
    ngx_http_request_t *r);
static void ngx_http_wasm_host_wasm_module_cleanup(void *data);
static ngx_http_wasm_module_state_t *ngx_http_wasm_host_get_module_ctx(
    ngx_http_request_t *r, ngx_http_wasm_host_rctx_t *rctx,
    ngx_http_wasm_module_t *hw_mod);


static ngx_int_t ngx_http_wasm_host_phase_handler(ngx_http_request_t *r,
    ngx_uint_t phase, const char *phase_name);
static ngx_int_t ngx_http_wasm_module_phase_handler(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod,
    ngx_uint_t phase);


static ngx_int_t ngx_http_wasm_post_read(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_server_rewrite(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_rewrite(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_preaccess(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_access(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_precontent(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_content(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_log(ngx_http_request_t *r);


static ngx_int_t ngx_http_wasm_host_postconf(ngx_conf_t *cf);
static void ngx_http_wasm_host_conf_cleanup(void *data);
static ngx_int_t ngx_http_wasm_host_init_worker(ngx_cycle_t *cycle);

static ngx_int_t ngx_http_wasm_host_init_instance(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_state_t *ctx);

static void ngx_http_wasm_host_body_handler(ngx_http_request_t *r);

static ngx_int_t ngx_http_wasm_response_body_filter_module(
    ngx_http_request_t *r, ngx_http_wasm_host_rctx_t *rctx,
    ngx_http_wasm_module_t *hw_mod, ngx_chain_t *in);
static ngx_int_t ngx_http_wasm_request_body_filter_module(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod,
    ngx_chain_t *in);
static ngx_int_t ngx_chain_add_copy_from(ngx_pool_t *pool, ngx_chain_t **chain,
    ngx_chain_t *in, ngx_chain_t **freep);
static ngx_int_t ngx_http_wasm_host_request_body_filter(ngx_http_request_t *r,
    ngx_chain_t *in);
static ngx_int_t ngx_http_wasm_host_header_filter(ngx_http_request_t *r);
static ngx_int_t ngx_http_wasm_host_response_body_filter(ngx_http_request_t *r,
    ngx_chain_t *in);


static char *ngx_http_wasm_host_variable(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static ngx_int_t ngx_http_wasm_host_variable_handler(ngx_http_request_t *r,
    ngx_http_variable_value_t *v, uintptr_t data);
static ngx_wasm_instance_t *ngx_http_wasm_host_variable_get_instance(
    ngx_http_request_t *r, ngx_wasm_module_t *wm);
static ngx_wasm_call_env_t *ngx_http_wasm_host_variable_create_call_env(
    ngx_http_request_t *r, ngx_http_wasm_host_var_ctx_t *ctx,
    ngx_str_t *result);


static void *ngx_http_wasm_host_create_main_conf(ngx_conf_t *cf);
static char *ngx_http_wasm_host_load_modules(ngx_conf_t *cf, void *conf);
static void *ngx_http_wasm_host_create_loc_conf(ngx_conf_t *cf);
static char *ngx_http_wasm_host_merge_loc_conf(ngx_conf_t *cf, void *ph,
    void *child);

static ngx_command_t  ngx_http_wasm_host_commands[] = {

    { ngx_string("wasm_header_filter"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
      ngx_http_wasm_header_filter,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_wasm_host_loc_conf_t, header_filters),
      NULL },

    { ngx_string("wasm_request_body_filter"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
      ngx_http_wasm_request_body_filter,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_wasm_host_loc_conf_t, request_body_filters),
      NULL },

    { ngx_string("wasm_response_body_filter"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
      ngx_http_wasm_response_body_filter,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_wasm_host_loc_conf_t, response_body_filters),
      NULL },

    { ngx_string("wasm_handler"),
      NGX_HTTP_LOC_CONF|NGX_CONF_NOARGS|NGX_CONF_1MORE,
      ngx_http_wasm_handler,
      NGX_HTTP_LOC_CONF_OFFSET,
      offsetof(ngx_http_wasm_host_loc_conf_t, phase_handlers),
      NULL },

    { ngx_string("wasm_env_argv"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_1MORE,
      ngx_http_wasm_env_argv,
      NGX_HTTP_LOC_CONF_OFFSET,
      0,
      NULL },

    { ngx_string("wasm_env_param"),
      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE2,
      ngx_http_wasm_env_param,
      NGX_HTTP_LOC_CONF_OFFSET,
      0,
      NULL },

    { ngx_string("wasm_var"),
      NGX_HTTP_MAIN_CONF|NGX_CONF_ANY,
      ngx_http_wasm_host_variable,
      NGX_HTTP_MAIN_CONF_OFFSET,
      0, NULL },

      ngx_null_command
};


static ngx_http_module_t  ngx_http_wasm_host_module_ctx = {
    NULL,                                  /* preconfiguration */
    ngx_http_wasm_host_postconf,           /* postconfiguration */

    ngx_http_wasm_host_create_main_conf,   /* create main configuration */
    NULL,                                  /* init main configuration */

    NULL,                                  /* create server configuration */
    NULL,                                  /* merge server configuration */

    ngx_http_wasm_host_create_loc_conf,     /* create location configuration */
    ngx_http_wasm_host_merge_loc_conf,      /* merge location configuration */
};


ngx_module_t  ngx_http_wasm_host_module = {
    NGX_MODULE_V1,
    &ngx_http_wasm_host_module_ctx,        /* module context */
    ngx_http_wasm_host_commands,           /* module directives */
    NGX_HTTP_MODULE,                       /* module type */
    NULL,                                  /* init master */
    NULL,                                  /* init module */
    ngx_http_wasm_host_init_worker,        /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    NULL,                                  /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};


static ngx_http_output_header_filter_pt  ngx_http_next_header_filter;
static ngx_http_request_body_filter_pt  ngx_http_next_request_body_filter;
static ngx_http_output_body_filter_pt  ngx_http_next_body_filter;

/* HTTP request apis */
static ngx_wasm_handle_ops_t ngx_http_wasm_apis[] = {
    {   .type = ngx_string("ngx::http::request"),
        .tag = NGX_WASM_HTTP_TAG_REQUEST,
        .open = ngx_wasm_http_request_open,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_http_request_property_get,
        .set = ngx_http_request_property_set,
        .close = ngx_wasm_http_request_close,
    },
    {   .type = ngx_string("ngx::http::request::headers"),
        .tag = NGX_WASM_HTTP_TAG_REQUEST_HEADERS,
        .open = ngx_wasm_http_open_request_headers,
        .close = ngx_wasm_host_close_stub,
        .read = ngx_wasm_http_read_request_headers,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
    },
    {   .type = ngx_string("ngx::http::request::body"),
        .tag = NGX_WASM_HTTP_TAG_REQUEST_BODY,
        .open = ngx_wasm_http_open_request_body,
        .close = ngx_wasm_host_close_stub,
        .read = ngx_wasm_http_read_request_body,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
    },
    {   .type = ngx_string("ngx::http::response::headers"),
        .tag = NGX_WASM_HTTP_TAG_RESPONSE_HEADERS,
        .open = ngx_wasm_http_open_response_headers,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_http_write_response_headers,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_http_set_response_headers,
        .close = ngx_wasm_host_close_stub
    },
    {   .type = ngx_string("ngx::http::response::body"),
        .tag = NGX_WASM_HTTP_TAG_RESPONSE_BODY,
        .open = ngx_wasm_http_open_response_body,
        .close = ngx_wasm_http_close_response_body,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_http_write_response_body,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
    },
    {   .type = ngx_string("ngx::http::location::env::kv"),
        .tag = NGX_WASM_HTTP_TAG_LOCATION_ENV_KV,
        .open = ngx_wasm_http_open_location_env_kv,
        .read = ngx_wasm_http_read_location_env_kv,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
        .close = ngx_wasm_host_close_stub,
    },
    {   .type = ngx_string("ngx::http::location::env::argv"),
        .tag = NGX_WASM_HTTP_TAG_LOCATION_ENV_ARGV,
        .open = ngx_wasm_http_open_location_env_argv,
        .read = ngx_wasm_http_read_location_env_argv,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
        .close = ngx_wasm_host_close_stub,
    },
#if (NGX_API)
    {   .type = ngx_string("ngx::api"),
        .tag = NGX_WASM_HOST_CALL_ENV_API_TAG,
        .open = ngx_wasm_host_open_api,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_api_get,
        .set = ngx_wasm_host_property_stub,
        .close = ngx_wasm_host_close_stub,
    },
#endif

};


static ngx_wasm_handle_ops_t ngx_http_wasm_conf_apis[] = {
    {   .type = ngx_string("ngx::http::config"),
        .tag = NGX_WASM_HTTP_TAG_CONFIG,
        .open = ngx_wasm_host_open_stub,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_http_host_set_conf,
        .close = ngx_wasm_host_close_stub,
    },

    ngx_wasm_host_null_ops
};


ngx_int_t
ngx_http_wasm_host_register_api(ngx_conf_t *cf, ngx_wasm_handle_ops_t *api)
{
    ngx_uint_t                        i;
    ngx_wasm_handle_ops_t           **ops, **opsp;
    ngx_http_wasm_host_main_conf_t   *hmcf;

    hmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_wasm_host_module);

    ops = hmcf->apis.elts;

    for (i = 0; i < hmcf->apis.nelts; i++) {
        if (ops[i]->tag == api->tag) {
            return NGX_OK;
        }
    }

    opsp = ngx_array_push(&hmcf->apis);
    if (opsp == NULL) {
        return NGX_ERROR;
    }

    *opsp = api;

    return NGX_OK;
}


static ngx_http_wasm_host_t *
ngx_http_wasm_host_create(ngx_wasm_instance_t *wi, ngx_http_request_t *r,
    const char *id)
{
    ngx_wasm_host_t                 *wh;
    ngx_http_wasm_host_t            *hh;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    wh = ngx_wasm_host_create(wi, r->pool, sizeof(ngx_http_wasm_host_t),
                              r->connection->log, id,
                              &hmcf->apis, wi->mod->acl);
    if (wh == NULL) {
        return NULL;
    }

    hh = (ngx_http_wasm_host_t *) wh;

    hh->request = r;

    return hh;
}


static int32_t
ngx_wasm_http_request_open(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    self->data = ngx_wasm_host_get_api_context(self, ph);
    if (self->data == NULL) {
        return NGX_ERROR;
    }

    return NGX_OK;
}


static ngx_inline int32_t
ngx_http_get_u32(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len,
    uint32_t value)
{

    if (len < sizeof(uint32_t)) {

        ngx_log_error(NGX_LOG_ERR, self->log, 0, "%V buffer is too small",
                      &self->host->name);
        return NGX_ERROR;
    }

    ngx_memcpy(buf, &value, sizeof(uint32_t));

    return sizeof(uint32_t);
}


static int32_t
ngx_http_request_property_get(ngx_wasm_handle_t *self, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    ngx_wasm_http_ctx_t *host_ctx = self->data;

    u_char              *p;
    ngx_log_t           *log;
    ngx_str_iter_t       si;
    ngx_http_request_t  *r;

    r = host_ctx->request;
    log = self->log;

    switch (propid) {

    case NGX_WASI_HTTP_REQUEST_PROPS_METHOD:
        return ngx_http_get_u32(self, buf, len, r->method);

    case NGX_WASI_HTTP_REQUEST_PROPS_SCHEMA:
        return ngx_http_get_u32(self, buf, len, ngx_wasm_http_get_schema(r));

    case NGX_WASI_HTTP_REQUEST_PROPS_VERSION:
        return ngx_http_get_u32(self, buf, len, r->http_version);

    case NGX_WASI_HTTP_REQUEST_PROPS_URI:

        ngx_iter_reset(&si);
        p = ngx_encode_str(&r->uri, &si, buf, buf + len);
        if (p == NULL) {
            return NGX_ERROR;
        }

        return p - buf;

    case NGX_WASI_HTTP_REQUEST_PROPS_REMOTE_ADDR:

        ngx_iter_reset(&si);
        p = ngx_encode_str(&r->connection->addr_text, &si, buf, buf + len);
        if (p == NULL) {
            return NGX_ERROR;
        }

        return p - buf;

    case NGX_WASI_HTTP_REQUEST_PROPS_VAR:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V get request variable", &self->host->name);

        return ngx_http_request_property_var(self->host, r, buf, len);

    default:

        ngx_log_error(NGX_LOG_ERR, log, 0,
                      "%V unknown request property %D",
                      &self->host->name, propid);

        return NGX_ERROR;
    }

    return 0;
}


static uint32_t
ngx_wasm_http_get_schema(ngx_http_request_t *r)
{
#if (NGX_HTTP_SSL)
    if (r->connection->ssl) {
        return 1;
    }
#endif

    return 0;
}


static int32_t
ngx_http_request_property_var(ngx_wasm_host_t *wh, ngx_http_request_t *r,
    uint8_t *buf, uint32_t len)
{
    uint32_t                   *in;
    ngx_uint_t                  hash;
    ngx_str_t                   key, val;
    ngx_http_variable_value_t  *vv;

    /*
     * input: array[4] of u32: [ key.len, key.ptr, val.len, val.ptr ]
     *
     * if val.len != 0 -> copy result, otherwise just return length
     */

    if (len != sizeof(uint32_t) * 4) {
        return NGX_ERROR;
    }

    in = (uint32_t *) buf;

    key.len = in[0];
    key.data = ngx_wasm_translate(wh->wi, (uintptr_t) in[1], key.len);

    val.len = in[2];
#if (NGX_SUPPRESS_WARN)
    val.data = NULL;
#endif

    if (val.len) {
        val.data = ngx_wasm_translate(wh->wi, (uintptr_t) in[3], val.len);
        if (val.data == NULL) {
            ngx_log_error(NGX_LOG_ERR, wh->log, 0,
                          "%V bad address", &wh->name);
            return NGX_ERROR;
        }
    }

    hash = ngx_hash_strlow(key.data, key.data, key.len);

    vv = ngx_http_get_variable(r, &key, hash);
    if (vv == NULL) {
        return NGX_ERROR;
    }

    if (vv->not_found) {
        return 0;
    }

    if (val.len == 0) {
        return vv->len;
    }

    if (vv->len > len) {
        ngx_log_error(NGX_LOG_ERR, wh->log, 0,
                      "%V buffer too small", &wh->name);
        return NGX_ERROR;
    }

    ngx_memcpy(val.data, vv->data, vv->len);

    return vv->len;
}


static int32_t
ngx_http_request_property_set(ngx_wasm_handle_t *self, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    ngx_wasm_http_ctx_t  *host_ctx = self->data;

    uint32_t             num;
    uint64_t             num64;
    ngx_int_t            rc;
    ngx_log_t           *log;
    ngx_http_request_t  *r;

    r = host_ctx->request;

    log = self->log;

    switch (propid) {

    case NGX_WASI_HTTP_REQUEST_PROPS_DISCARD_BODY:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V set request discard body", &self->host->name);

        if (len < sizeof(uint8_t)) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument", &self->host->name);
            return NGX_ERROR;
        }

        if (buf[0] != 1) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument value", &self->host->name);
            return NGX_ERROR;
        }

        rc = ngx_http_discard_request_body(r);
        if (rc != NGX_OK) {
            return NGX_ERROR;
        }

        return len;

    case NGX_WASI_HTTP_REQUEST_PROPS_STATUS:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V set request response status", &self->host->name);

        if (len < sizeof(uint32_t)) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument", &self->host->name);
            return NGX_ERROR;
        }

        ngx_memcpy(&num, buf, sizeof(uint32_t));

        /* TODO: check status for validity */

        r->headers_out.status = num;

        return len;

    case NGX_WASI_HTTP_REQUEST_PROPS_REJECT_MSG:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V set request response reject message",
                       &self->host->name);

        if (len > 2048) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument", &self->host->name);
            return NGX_ERROR;
        }

        host_ctx->reject_msg.data = ngx_pnalloc(r->pool, len);
        if (host_ctx->reject_msg.data == NULL) {
            return NGX_ERROR;
        }

        ngx_memcpy(host_ctx->reject_msg.data, buf, len);
        host_ctx->reject_msg.len = len;

        return len;

    case NGX_WASI_HTTP_REQUEST_PROPS_CONTENT_LENGTH:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V set request content length", &self->host->name);

        if (len < sizeof(uint64_t)) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument", &self->host->name);
            return NGX_ERROR;
        }

        ngx_memcpy(&num64, buf, sizeof(uint64_t));

        if (num64 == (uint64_t) -1) {
            /* chunked  */
            r->headers_out.content_length_n = -1;

        } else {
            r->headers_out.content_length_n = num64;
        }

        return len;

    case NGX_WASI_HTTP_REQUEST_PROPS_HEADER_ONLY:
        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, log, 0,
                       "%V set request header only", &self->host->name);

        if (len < sizeof(uint8_t)) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument", &self->host->name);
            return NGX_ERROR;
        }

        if (buf[0] == 0) {
            r->header_only = 0;

        } else if (buf[0] == 1) {
            r->header_only = 1;

        } else{
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V invalid argument value", &self->host->name);
            return NGX_ERROR;
        }

        return len;

    default:

        ngx_log_error(NGX_LOG_ERR, log, 0,
                      "%V set request unknown property id: %D",
                      &self->host->name, propid);

        return NGX_ERROR;
    }

    return 0;
}


static ngx_int_t
ngx_http_wasm_host_reject_request(ngx_http_request_t *r,
    ngx_str_t *reject_msg)
{
    ngx_int_t    rc;
    ngx_chain_t  out;

    /* TODO: mark state as dead ? */

    ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "http wasm content request rejected");

    if (r->header_sent) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                      "http wasm content failed to reject request: "
                      "headers already sent");
        return NGX_ERROR;
    }

    if (reject_msg->len) {
        out.buf = ngx_calloc_buf(r->pool);
        if (out.buf == NULL) {
            return NGX_ERROR;
        }

        out.buf->start = reject_msg->data;
        out.buf->last = out.buf->start + reject_msg->len;
        out.buf->end = out.buf->last;
        out.buf->pos = out.buf->start;
        out.buf->memory = 1;

        out.buf->last_buf = (r == r->main) ? 1 : 0;
        out.buf->last_in_chain = 1;

        r->headers_out.content_length_n = out.buf->last - out.buf->pos;

    } else {
        r->headers_out.content_length_n = 0;
    }

    rc = ngx_http_send_header(r);

    if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {
        return rc;
    }

    out.next = NULL;

    if (reject_msg->len) {
        return ngx_http_output_filter(r, &out);

    } else {
        return ngx_http_output_filter(r, NULL);
    }
}


static int32_t
ngx_wasm_http_request_close(ngx_wasm_handle_t *self)
{
    return NGX_OK;
}


static int32_t
ngx_http_wasm_host_get_curr_header(uint64_t i, void *data, void *elt)
{
    ngx_kv_t                     *kv;
    ngx_table_elt_t              *te;
    ngx_wasm_http_headers_pos_t  *list_pos;

    kv = elt;

    list_pos = data;

    te = list_pos->te;

    kv->key = te->key;
    kv->value = te->value;

    return NGX_OK;
}


static int32_t
ngx_http_wasm_host_next_header(void *data)
{
    ngx_wasm_http_headers_pos_t  *pos;

    pos = data;

    ngx_table_elt_t  *items;
    ngx_list_part_t  *part;

    part = pos->part;

    pos->off++;

    if (pos->off >= part->nelts) {
        if (part->next == NULL) {
            return NGX_ERROR;
        }
        part = part->next;

        pos->part = part;
        pos->off = 0;
    }

    items = part->elts;

    pos->te = &items[pos->off];

    return NGX_OK;
}


static void
ngx_wasm_http_init_request_headers_encoder(ngx_kv_array_t *kva,
    ngx_list_part_t *part, ngx_wasm_http_headers_pos_t *pos)
{
    ngx_uint_t        i;
    ngx_table_elt_t  *headers;

    ngx_kv_array_init(kva, 0);

    pos->part = part;
    pos->off = 0;
    pos->te = part->elts;

    headers = part->elts;

    for (i = 0; /* void */ ; i++) {
        if (i >= part->nelts) {
            if (part->next == NULL) {
                break;
            }
            part = part->next;
            headers = part->elts;
            i = 0;
        }

        kva->arr.nelts++;
        kva->arr.size += headers[i].key.len + headers[i].value.len;
    }

    kva->arr.data = pos;

    kva->arr.curr = ngx_http_wasm_host_get_curr_header;
    kva->arr.next = ngx_http_wasm_host_next_header;
}


static int32_t
ngx_wasm_http_open_request_headers(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_http_request_t               *r;
    ngx_wasm_http_ctx_t              *host_ctx;
    ngx_http_open_request_headers_t  *orh;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V open request headers", &self->host->name);

    host_ctx = ngx_wasm_host_get_object_context(self, ph,
                                                NGX_WASM_HTTP_TAG_REQUEST);
    if (host_ctx == NULL) {
        return NGX_ERROR;
    }

    r = host_ctx->request;

    orh = ngx_palloc(self->host->pool, sizeof(ngx_http_open_request_headers_t));
    if (orh == NULL) {
        return NGX_ERROR;
    }

    ngx_wasm_http_init_request_headers_encoder(&orh->kva,
                                               &r->headers_in.headers.part,
                                               &orh->pos);

    self->data = orh;

    return NGX_OK;
}


static int32_t
ngx_wasm_http_read_request_headers(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    uint8_t  *p;
    ngx_http_open_request_headers_t  *orh;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V read request headers", &self->host->name);

    orh = self->data;

    p = ngx_encode_array(&orh->kva.arr, &orh->kva.ai, buf, buf + len);

    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


static int32_t
ngx_wasm_http_open_request_body(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_http_request_t            *r;
    ngx_wasm_http_ctx_t           *host_ctx;
    ngx_http_open_request_body_t  *rqb;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V open request body", &self->host->name);

    host_ctx = ngx_wasm_host_get_object_context(self, ph,
                                                NGX_WASM_HTTP_TAG_REQUEST);
    if (host_ctx == NULL) {
        return NGX_ERROR;
    }

    r = host_ctx->request;

    rqb = ngx_palloc(self->host->pool, sizeof(ngx_http_open_request_body_t));
    if (rqb == NULL) {
        return NGX_ERROR;
    }

    rqb->r = r;
    rqb->body = NULL; /* r->request_body is not set yet */
    rqb->offset = 0;
    rqb->read_started = 0;
    rqb->host_ctx = host_ctx;

    self->data = rqb;

    return NGX_OK;
}


static int32_t
ngx_wasm_http_read_request_body(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    ngx_http_open_request_body_t *rqb = self->data;

    ngx_http_wasm_host_rctx_t *rctx;

    size_t      n;
    ngx_int_t   rc;
    ngx_log_t  *log;

    log = self->log;

    rctx = (ngx_http_wasm_host_rctx_t *) ngx_http_get_module_ctx(rqb->r,
                                                    ngx_http_wasm_host_module);

    ngx_log_debug3(NGX_LOG_DEBUG_HTTP, log, 0,
                   "%V open read request body rs:%i isc:%d",
                   &self->host->name, rqb->read_started,
                   rqb->host_ctx->in_syscall);


    if (!rqb->read_started) {

        if (rctx->http_state.body_handler) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V attempt to read body by different modules",
                          &self->host->name);
            return NGX_ERROR;
        }

        /* native handler for body post-read */
        rctx->http_state.body_handler = ngx_http_wasm_host_body_handler;
        /* wasm function to enter from native handler */
        rctx->http_state.body_handler_params =
                         &rctx->curr_module->hw->phase_hooks[rctx->curr_phase];

        /* protect from calling wasm again from post handler or filter */
        rqb->host_ctx->in_syscall = 1;

        rc = ngx_http_read_client_request_body(rqb->r,
                                               rqb->host_ctx->body_handler);
        rqb->host_ctx->in_syscall = 0;

        if (rc >= NGX_HTTP_SPECIAL_RESPONSE) {
            ngx_log_error(NGX_LOG_ERR, log, 0,
                          "%V internal server error: "
                          "special response %i", &self->host->name, rc);
            return NGX_ERROR;
        }

        rqb->read_started = 1;

        if (!rqb->host_ctx->post_body_called) {
            /* the body_handler will invoke wasm handler later */
            return NGX_DONE;
        }

        /* else: body can be read, proceed further  */
    }

    if (rqb->r->request_body == NULL) {
        ngx_log_error(NGX_LOG_ERR, log, 0,
                      "%V handler called with no body", &self->host->name);
        return NGX_ERROR;
    }

    rqb->body = rqb->r->request_body;

    /* wasm guest expects body, but client did not send it  */
    if (rqb->body->bufs == NULL) {
        ngx_log_error(NGX_LOG_ERR, log, 0,
                      "%V handler called with empty body", &self->host->name);
        return NGX_ERROR;
    }

    n = 0;

    if (rqb->body->temp_file) {
        rc = ngx_wasm_http_body_from_file(rqb->body, rqb->offset, buf, len,
                                             &n, log);

    } else {
        rc = ngx_wasm_http_body_from_mem(rqb->body, rqb->offset, buf, len, &n);
    }

    if (rc != NGX_OK) {
        return NGX_ERROR;
    }

    rqb->offset += n;

    return n;
}


static ngx_int_t
ngx_wasm_http_body_from_mem(ngx_http_request_body_t *body, off_t off,
    uint8_t *buf, size_t size, size_t *nread)
{
    size_t        n, left, len;
    uint8_t      *pos, *rpos,skip;
    ngx_buf_t    *b;
    ngx_chain_t  *cl;

    n = 0;
    pos = buf;

    if (off) {
        skip = 1;
        left = off;

    } else {
        skip = 0;
        left = size;
    }

    for (cl = body->bufs; cl; cl = cl->next) {

        b = cl->buf;

        if (ngx_buf_special(b)) {
            continue;
        }

        len = ngx_buf_size(b);

        if (len < left) {

            left -= len;

            if (!skip) {
                pos = ngx_cpymem(pos, b->pos, len);
                n += len;
            }

            continue;
        }

        /* len >= left */

        if (skip) {
            skip = 0;
            len -= left;
            rpos = b->pos +left;

            left = size;

            if (len < left) {
                left -= len;

                ngx_memcpy(pos, rpos, len);
                n += len;

                continue;

            } else {

                ngx_memcpy(pos, rpos, left);
                n += left;

                break;
            }

        } else {

            ngx_memcpy(pos, b->pos, left);
            n += left;

            break;
        }
    }

    *nread = n;

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_http_body_from_file(ngx_http_request_body_t *body, off_t off,
    uint8_t *buf, size_t size, size_t *nread, ngx_log_t *log)
{
    int      fd;
    ssize_t  n;

    fd = body->temp_file->file.fd;

    if (lseek(fd, off, SEEK_SET) == -1) {
        ngx_log_error(NGX_LOG_ERR, log, ngx_errno,
                      "wasm http body from file lseek(%D, %O) failed",
                      fd, off);
        return NGX_ERROR;
    }

    n = ngx_read_fd(fd, buf, size);
    if (n == -1) {
        ngx_log_error(NGX_LOG_ERR, log, ngx_errno,
                      "wasm http body from file read(%D, %z) failed",
                      fd, size);
        return NGX_ERROR;
    }

    *nread = (size_t) n;

    return NGX_OK;
}


static int32_t
ngx_http_wasm_init_response_header(ngx_arr_t *arr, uint64_t off, void *elt)
{
    ngx_http_request_t *r = arr->data;

    ngx_kv_t         *kv;
    ngx_table_elt_t  *h;

    kv = elt;

    h = ngx_list_push(&r->headers_out.headers);
    if (h == NULL) {
        return NGX_ERROR;
    }

    h->hash = 1;
    h->key = kv->key;
    h->value = kv->value;

    return NGX_OK;
}


static int32_t
ngx_wasm_http_open_response_headers(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_wasm_http_ctx_t               *host_ctx;
    ngx_http_open_response_headers_t  *ctx;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V open response headers", &self->host->name);

    host_ctx = ngx_wasm_host_get_object_context(self, ph,
                                                NGX_WASM_HTTP_TAG_REQUEST);
    if (host_ctx == NULL) {
        return NGX_ERROR;
    }

    ctx = ngx_pcalloc(self->host->pool,
                      sizeof(ngx_http_open_response_headers_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ngx_kv_array_init(&ctx->kva, 0);

    ctx->kva.arr.init = ngx_http_wasm_init_response_header;
    ctx->kva.arr.data = host_ctx->request;
    ctx->kva.ac.pool = host_ctx->request->pool;

    self->data = ctx;

    return NGX_OK;
}



static int32_t
ngx_wasm_http_write_response_headers(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    ngx_http_open_response_headers_t *ctx = self->data;

    uint8_t  *p;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V write response headers", &self->host->name);

    p = ngx_decode_array(&ctx->kva.arr, &ctx->kva.ai, buf, buf + len, &ctx->kva.ac);
    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


static int32_t
ngx_wasm_http_set_response_headers(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len)
{
    ngx_http_open_response_headers_t *ctx = self->data;

    ngx_int_t            rc;
    ngx_http_request_t  *r;

    if (propid != NGX_WASI_HTTP_REQUEST_HEADER_OPS_SEND) {
        ngx_log_debug2(NGX_LOG_DEBUG_HTTP, self->log, 0,
                       "%V set response headers unknown property %D",
                       &self->host->name, propid);
        return NGX_ERROR;
    }

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V send response headers", &self->host->name);

    r = ctx->kva.arr.data;

    rc = ngx_http_send_header(r);

    if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {

        if (rc > NGX_OK) {
            ngx_log_error(NGX_LOG_ERR, self->log, 0,
                          "%V ngx_http_send_header() failed with %i",
                          &self->host->name, rc);
            rc = NGX_ERROR;
        }

        return rc;
    }

    return NGX_OK;
}


static int32_t
ngx_wasm_http_open_response_body(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_wasm_http_ctx_t            *host_ctx;
    ngx_http_open_response_body_t  *rspb;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V open response body", &self->host->name);

    host_ctx = ngx_wasm_host_get_object_context(self, ph,
                                                NGX_WASM_HTTP_TAG_REQUEST);
    if (host_ctx == NULL) {
        return NGX_ERROR;
    }

    rspb = ngx_pcalloc(self->host->pool, sizeof(ngx_http_open_response_body_t));
    if (rspb == NULL) {
        return NGX_ERROR;
    }

    rspb->r = host_ctx->request;

    self->data = rspb;

    return NGX_OK;
}


static int32_t
ngx_wasm_http_write_response_body(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t bytes)
{
    ngx_buf_t    *b;
    ngx_chain_t  *cl, **ll;

    ngx_http_open_response_body_t  *rspb;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V write response body", &self->host->name);

    if (bytes == 0) {
        return 0;
    }

    rspb = self->data;

    for (cl = rspb->out_bufs, ll = &rspb->out_bufs; cl; cl = cl->next) {
        ll = &cl->next;
    }

    cl = ngx_chain_get_free_buf(self->host->pool, &rspb->free_bufs);
    if (cl == NULL) {
        return NGX_ERROR;
    }

    *ll = cl;

    b = cl->buf;

    // TODO: check if need add CRLF to end (test with chunked encoding?)
    //b->pos = ngx_palloc(self->host->pool, bytes + 2);
    b->pos = ngx_palloc(self->host->pool, bytes);
    if (b->pos == NULL) {
        return NGX_ERROR;
    }

    b->flush = 1;
    b->memory = 1;
    ngx_memcpy(b->pos, buf, bytes);
    b->last = b->pos + bytes;
//    *b->last++ = CR;
//    *b->last++ = LF;

    return bytes;
}


static int32_t
ngx_wasm_http_close_response_body(ngx_wasm_handle_t *self)
{
    ngx_int_t                       rc;
    ngx_buf_t                      *b;
    ngx_chain_t                    *cl, **ll;
    ngx_http_request_t             *r;
    ngx_http_open_response_body_t  *rspb;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, self->log, 0,
                   "%V close response body", &self->host->name);

    rspb = self->data;
    r = rspb->r;

    for (cl = rspb->out_bufs, ll = &rspb->out_bufs; cl; cl = cl->next) {
        ll = &cl->next;
    }

    cl = ngx_chain_get_free_buf(self->host->pool, &rspb->free_bufs);
    if (cl == NULL) {
        return NGX_ERROR;
    }

    *ll = cl;

    b = cl->buf;

    b->flush = 1;
    b->last_in_chain = 1;
    b->last_buf = (r == r->main) ? 1 : 0;

    rc = ngx_http_output_filter(rspb->r, rspb->out_bufs);

    if (rc == NGX_ERROR || rc > NGX_OK) {

        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V ngx_http_output_filter() failed with %i",
                      &self->host->name, rc);

        return NGX_ERROR;
    }

    return NGX_OK;
}


static char *
ngx_http_wasm_env_argv(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    ngx_str_t                         *value;
    ngx_uint_t                         i;
    ngx_http_complex_value_t          *arg;
    ngx_http_compile_complex_value_t   ccv;

    if (wfcf->env.argv != NGX_CONF_UNSET_PTR) {
        return "duplicate";
    }

    wfcf->env.argv = ngx_array_create(cf->pool, cf->args->nelts,
                                   sizeof(ngx_http_complex_value_t));
    if (wfcf->env.argv == NULL) {
        return NGX_CONF_ERROR;
    }

    value = cf->args->elts;

    for (i = 1; i < cf->args->nelts; i++) {

        arg = ngx_array_push(wfcf->env.argv);
        if (arg == NULL) {
            return NGX_CONF_ERROR;
        }

        ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));

        ccv.cf = cf;
        ccv.value = &value[i];
        ccv.complex_value = arg;

        if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
            return NGX_CONF_ERROR;
        }
    }

    return NGX_CONF_OK;
}


static char *
ngx_http_wasm_env_param(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    ngx_str_t                         *value;
    ngx_http_wasm_env_param_t         *param;
    ngx_http_compile_complex_value_t   ccv;

    if (wfcf->env.env_params == NGX_CONF_UNSET_PTR) {

        wfcf->env.env_params = ngx_array_create(cf->pool, 4,
                                       sizeof(ngx_http_wasm_env_param_t));
        if (wfcf->env.env_params == NULL) {
            return NGX_CONF_ERROR;
        }
    }

    value = cf->args->elts;

    param = ngx_array_push(wfcf->env.env_params);
    if (param == NULL) {
        return NGX_CONF_ERROR;
    }

    ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));

    ccv.cf = cf;
    ccv.value = &value[1];
    ccv.complex_value = &param->key;

    if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
        return NGX_CONF_ERROR;
    }

    ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));

    ccv.cf = cf;
    ccv.value = &value[2];
    ccv.complex_value = &param->value;

    if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


static ngx_http_wasm_module_t *
ngx_http_wasm_host_find_module(ngx_conf_t *cf, ngx_str_t *name)
{
    ngx_uint_t                       i;
    ngx_http_wasm_module_t          *hw_modules;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    hmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_wasm_host_module);
    hw_modules = hmcf->modules.elts;

    for (i = 0; i < hmcf->modules.nelts; i++) {
        if (hw_modules[i].name.len == name->len
            && ngx_strncmp(hw_modules[i].name.data, name->data, name->len)
               == 0)
        {
            return &hw_modules[i];
        }
    }

    return NULL;
}


static char *
ngx_http_wasm_add_filter_handler(ngx_conf_t *cf, ngx_command_t *cmd,
    ngx_array_t **handlers)
{
    ngx_str_t                       *value;
    ngx_uint_t                      *idx;
    ngx_http_wasm_module_t          *hw_mod;

    value = cf->args->elts;
    value = &value[1];

    if (value->len == 3 && ngx_strncmp(value->data, "off", 3) == 0) {

        if (*handlers != NULL && *handlers != NGX_CONF_UNSET_PTR) {
            return "attempt to mix special \"off\" value with handlers";
        }

        *handlers = NULL;
        return NGX_CONF_OK;

    }

    if (*handlers == NULL) {
        return "attempt to mix special \"off\" value with handlers";
    }

    hw_mod = ngx_http_wasm_host_find_module(cf, value);
    if (hw_mod == NULL) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "unknown http wasm module \"%V\"", value);
        return NGX_CONF_ERROR;
    }

    if (*handlers == NGX_CONF_UNSET_PTR) {
        *handlers = ngx_array_create(cf->pool, 4, sizeof(ngx_uint_t));
        if (*handlers == NULL) {
            return NGX_CONF_ERROR;
        }
    }

    idx = ngx_array_push(*handlers);
    if (idx == NULL) {
        return NGX_CONF_ERROR;
    }

    *idx = hw_mod->own_index;

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_http_wasm_add_phase_handler(ngx_conf_t *cf, ngx_uint_t phase,
    ngx_array_t **handlers, ngx_uint_t mod_index)
{
    ngx_uint_t  *idx;

    if (*handlers == NULL) {
        *handlers = ngx_array_create(cf->pool, 4, sizeof(ngx_uint_t));
        if (*handlers == NULL) {
            return NGX_ERROR;
        }
    }

    if (phase == NGX_HTTP_CONTENT_PHASE && (*handlers)->nelts) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "duplicate http wasm content handler in location");
        return NGX_ERROR;
    }

    idx = ngx_array_push(*handlers);
    if (idx == NULL) {
        return NGX_ERROR;
    }

    *idx = mod_index;

    return NGX_OK;
}


static char *
ngx_http_wasm_handler(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    ngx_str_t                *value;
    ngx_uint_t                i;
    ngx_array_t             **handlers /* per phase */;
    ngx_http_wasm_module_t   *hw_mod;

    value = cf->args->elts;

    if (value[1].len == 3 && ngx_strncmp(value[1].data, "off", 3) == 0) {

        if (wfcf->phase_handlers != NULL
            && wfcf->phase_handlers != NGX_CONF_UNSET_PTR)
        {
            return "attempt to mix special \"off\" value with phase handlers";
        }

        wfcf->phase_handlers = NULL;

        return NGX_CONF_OK;

    }

    if (wfcf->phase_handlers == NULL) {
        return "attempt to mix special \"off\" value with phase handlers";
    }

    hw_mod = ngx_http_wasm_host_find_module(cf, &value[1]);
    if (hw_mod == NULL) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "unknown http wasm module \"%V\"", &value[1]);
        return NGX_CONF_ERROR;
    }

    if (wfcf->phase_handlers == NGX_CONF_UNSET_PTR) {
        wfcf->phase_handlers = ngx_pcalloc(cf->pool,
                             sizeof(ngx_array_t *) * (NGX_HTTP_LOG_PHASE + 1));
        if (wfcf->phase_handlers == NULL) {
            return NGX_CONF_ERROR;
        }
    }

    if (cf->args->nelts < 3) {

        /* handle all phases by this module in this location */

        for (i = 0; i < NGX_HTTP_LOG_PHASE + 1; i++) {

            handlers = &((*wfcf->phase_handlers)[i]);

            if (ngx_http_wasm_add_phase_handler(cf, i,
                                                handlers, hw_mod->own_index)
                != NGX_OK)
            {
                return NGX_CONF_ERROR;
            }
        }

        return NGX_CONF_OK;
    }

#define ngx_http_wasm_host_phase_bind(name, phase)                            \
    if (value[i].len == (sizeof(name) - 1)                                    \
        && ngx_strncmp(value[i].data, name, sizeof(name) - 1) == 0)           \
    {                                                                         \
        handlers = &((*wfcf->phase_handlers)[phase]);                         \
        if (ngx_http_wasm_add_phase_handler(cf, phase, handlers,              \
                                            hw_mod->own_index)                \
            != NGX_OK)                                                        \
        {                                                                     \
            return NGX_CONF_ERROR;                                            \
        }                                                                     \
        continue;                                                             \
    }

    /* phases are specified explicitely */

    for (i = 2; i < cf->args->nelts; i++) {

        ngx_http_wasm_host_phase_bind("post_read", NGX_HTTP_POST_READ_PHASE);
        ngx_http_wasm_host_phase_bind("server_rewrite",
                                       NGX_HTTP_SERVER_REWRITE_PHASE);
        ngx_http_wasm_host_phase_bind("rewrite", NGX_HTTP_REWRITE_PHASE);
        ngx_http_wasm_host_phase_bind("preaccess", NGX_HTTP_PREACCESS_PHASE);
        ngx_http_wasm_host_phase_bind("access", NGX_HTTP_ACCESS_PHASE);
        ngx_http_wasm_host_phase_bind("precontent", NGX_HTTP_PRECONTENT_PHASE);
        ngx_http_wasm_host_phase_bind("content", NGX_HTTP_CONTENT_PHASE);
        ngx_http_wasm_host_phase_bind("log", NGX_HTTP_LOG_PHASE);

        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "unknown http phase \"%V\"", &value[i]);

        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


static char *
ngx_http_wasm_header_filter(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    return ngx_http_wasm_add_filter_handler(cf, cmd, &wfcf->header_filters);
}


static char *
ngx_http_wasm_request_body_filter(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    return ngx_http_wasm_add_filter_handler(cf, cmd,
                                            &wfcf->request_body_filters);
}


static char *
ngx_http_wasm_response_body_filter(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf)
{
    ngx_http_wasm_host_loc_conf_t  *wfcf = conf;

    return ngx_http_wasm_add_filter_handler(cf, cmd,
                                            &wfcf->response_body_filters);
}


static ngx_array_t *
ngx_http_wasm_host_create_call_env_kv(ngx_http_request_t *r)
{
    ngx_uint_t                  i;
    ngx_kv_t                   *kv;
    ngx_array_t                *ret;
    ngx_http_wasm_env_ctx_t    *env;
    ngx_http_wasm_env_param_t  *params;

    env = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (env == NULL || env->env_params == NULL) {
        return ngx_array_create(r->pool, 1, sizeof(ngx_kv_t));
    }

    i = ngx_max(1, env->env_params->nelts);

    ret = ngx_array_create(r->pool, env->env_params->nelts, sizeof(ngx_kv_t));
    if (ret == NULL) {
        return NULL;
    }

    params = env->env_params->elts;

    for (i = 0; i < env->env_params->nelts; i++) {

        kv = ngx_array_push(ret);

        if (ngx_http_complex_value(r, &params[i].key, &kv->key) != NGX_OK) {
            return NULL;
        }

        if (ngx_http_complex_value(r, &params[i].value, &kv->value) != NGX_OK)
        {
            return NULL;
        }
    }

    return ret;
}


static ngx_array_t *
ngx_http_wasm_host_create_call_env_argv(ngx_http_request_t *r)
{
    ngx_str_t                 *str;
    ngx_uint_t                 i;
    ngx_array_t               *ret;
    ngx_http_wasm_env_ctx_t   *env;
    ngx_http_complex_value_t  *cvs;

    env = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (env == NULL || env->argv == NULL) {
        return ngx_array_create(r->pool, 1, sizeof(ngx_str_t));
    }

    i = ngx_max(1, env->argv->nelts);

    ret = ngx_array_create(r->pool, i, sizeof(ngx_str_t));
    if (ret == NULL) {
        return NULL;
    }

    cvs = env->argv->elts;

    for (i = 0; i < env->argv->nelts; i++) {

        str = ngx_array_push(ret);

        if (ngx_http_complex_value(r, &cvs[i], str) != NGX_OK) {
            return NULL;
        }
    }

    return ret;
}


static int32_t
ngx_wasm_http_open_location_env_kv(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_http_wasm_env_state_t  *ctx;

    ngx_http_wasm_host_t *hh = (ngx_http_wasm_host_t *) self->host;

    if (hh->ce_kvs == NULL) {
        hh->ce_kvs = ngx_http_wasm_host_create_call_env_kv(hh->request);
        if (hh->ce_kvs == NULL) {
            return NGX_ERROR;
        }
    }

    ctx = ngx_pcalloc(self->host->pool, sizeof(ngx_http_wasm_env_state_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ngx_kv_array_init_array_encoder(&ctx->kva, hh->ce_kvs->nelts,
                                    hh->ce_kvs->elts);
    self->data = ctx;

    return NGX_OK;
}


static int32_t
ngx_wasm_http_open_location_env_argv(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph)
{
    ngx_http_wasm_env_state_t  *ctx;

    ngx_http_wasm_host_t *hh = (ngx_http_wasm_host_t *) self->host;

    if (hh->ce_argv == NULL) {
        hh->ce_argv = ngx_http_wasm_host_create_call_env_argv(hh->request);
        if (hh->ce_argv == NULL) {
            return NGX_ERROR;
        }
    }

    ctx = ngx_pcalloc(self->host->pool, sizeof(ngx_http_wasm_env_state_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ngx_str_array_init_array_encoder(&ctx->sa, hh->ce_argv->nelts,
                                     hh->ce_argv->elts);
    self->data = ctx;

    return NGX_OK;

}


static int32_t
ngx_wasm_http_read_location_env_kv(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    ngx_http_wasm_env_state_t  *ce = self->data;

    uint8_t  *p;

    p = ngx_encode_array(&ce->kva.arr, &ce->kva.ai, buf, buf + len);
    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


static int32_t
ngx_wasm_http_read_location_env_argv(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    ngx_http_wasm_env_state_t *ce = self->data;

    uint8_t  *p;

    p = ngx_encode_array(&ce->sa.arr, &ce->sa.ai, buf, buf + len);
    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


#if (NGX_API)

static int32_t
ngx_wasm_host_open_api(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    ngx_wasm_api_ctx_t    *ac;
    ngx_http_wasm_host_t  *hh;

    hh = (ngx_http_wasm_host_t *) self->host;

    ac = ngx_pcalloc(self->host->pool, sizeof(ngx_wasm_api_ctx_t));
    if (ac == NULL) {
        return NGX_ERROR;
    }

    ac->result_iter.state = NGX_ITER_ST_FINAL;

    ac->root = ngx_api_root((ngx_cycle_t *) ngx_cycle);

    /* TODO: remove connection from API context */
    ac->connection = hh->request->connection;

    self->data = ac;

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_host_api_get_result(ngx_wasm_api_ctx_t *ac, ngx_pool_t *pool,
    ngx_str_t *path, ngx_str_t *res)
{
    ngx_int_t      rc;
    ngx_chain_t    out;
    ngx_api_ctx_t  ctx;

    ngx_memzero(&ctx, sizeof(ngx_api_ctx_t));

    ctx.connection = ac->connection;
    ctx.pool = pool;
    ctx.config_files = 0;
    ctx.pretty = 1;
    ctx.epoch = 0;
    ctx.path = *path;
    ctx.orig_path = *path;

    rc = ac->root->handler(ac->root->data, &ctx, NULL);

    if (rc == NGX_OK && ctx.path.len) {
        rc = NGX_API_NOT_FOUND;
    }

    if (rc == NGX_DECLINED) {
        rc = NGX_API_NOT_FOUND;
    }

    switch (rc) {

    case NGX_ERROR:
        return NGX_ERROR;

    case NGX_API_NOT_FOUND:
        return NGX_DECLINED;
    }

    /* NGX_OK */

    out.buf = ngx_json_render(pool, ctx.out, ctx.pretty);
    if (out.buf == NULL) {
        return NGX_ERROR;
    }

    res->data = out.buf->pos;
    res->len = out.buf->last - out.buf->pos;

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_host_api_decode_key(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len, ngx_str_t *res)
{
    u_char          *p;
    uint64_t         slen;
    ngx_len_iter_t   iter;

    ngx_iter_reset(&iter);
    p = ngx_decode_len(&slen, &iter, buf, buf + len);

    if (!ngx_iter_done(&iter)) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0, "%V failed to get key length",
                      &self->host->name);
        return NGX_ERROR;
    }

    len -= (p - buf);

    if (slen > len) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0, "%V incomplete key passed",
                      &self->host->name);
        return NGX_ERROR;
    }

    res->len = slen;
    res->data = p;

    return NGX_OK;
}


static int32_t
ngx_wasm_host_api_get(ngx_wasm_handle_t *self, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    ngx_wasm_api_ctx_t *ac = self->data;

    uint8_t    *p;
    ngx_int_t   rc;
    ngx_str_t   key;

    if (ac->api_result.data && !ngx_iter_done(&ac->result_iter)) {
        /* continuation */

        p = ngx_encode_str(&ac->api_result, &ac->result_iter, buf, buf + len);
        if (p == NULL) {
            ngx_log_error(NGX_LOG_ERR, self->log, 0,
                          "%V failed encode api get result on continue",
                          &self->host->name);
            return NGX_ERROR;
        }

        return p - buf;
    }

    rc = ngx_wasm_host_api_decode_key(self, buf, len, &key);
    if (rc != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V failed to decode api get key",
                      &self->host->name);
        return rc;
    }

    rc = ngx_wasm_host_api_get_result(ac, self->host->pool, &key,
                                      &ac->api_result);
    if (rc != NGX_OK) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V failed to obtain result from API",
                      &self->host->name);
        return rc;
    }

    ngx_iter_reset(&ac->result_iter);

    p = ngx_encode_str(&ac->api_result, &ac->result_iter, buf, buf + len);
    if (p == NULL) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V failed to encode api get result",
                      &self->host->name);
        return NGX_ERROR;
    }

    if (ngx_iter_done(&ac->result_iter)) {
        ngx_iter_reset(&ac->result_iter);
    }

    return p - buf;
}

#endif


static ngx_int_t
ngx_http_wasm_host_add_ops(ngx_array_t *apis, ngx_wasm_handle_ops_t *ops)
{
    ngx_wasm_handle_ops_t  **p;

    while (ops->type.data) {
        p = ngx_array_push(apis);
        if (p == NULL) {
            return NGX_ERROR;
        }

        *p = ops++;
    }

    return NGX_OK;
}


static ngx_http_wasm_host_conf_t *
ngx_http_wasm_conf_host_create(ngx_wasm_instance_t *wi, const char *id)
{
    ngx_array_t                *apis;
    ngx_wasm_host_t            *wh;
    ngx_http_wasm_host_conf_t  *whcf;

    apis = ngx_array_create(wi->pool, 2, sizeof(ngx_wasm_handle_ops_t *));
    if (apis == NULL) {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(apis, ngx_http_wasm_conf_apis) != NGX_OK) {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(apis, ngx_wasm_log_ops) != NGX_OK) {
        return NULL;
    }

    wh = ngx_wasm_host_create(wi, wi->pool, sizeof(ngx_http_wasm_host_conf_t),
                              wi->log, id, apis, NULL);
    if (wh == NULL) {
        return NULL;
    }

    whcf = (ngx_http_wasm_host_conf_t *) wh;

    whcf->conf_fd = ngx_wasm_host_create_object(wh, "ngx::http::config",
                                                wi->log);
    if (whcf->conf_fd == -1) {
        return NULL;
    }

    return whcf;
}


static ngx_int_t
ngx_wasm_http_host_install_phase_hook(ngx_http_wasm_host_conf_t *host,
    ngx_uint_t phase, const char *phase_name, uint8_t *buf, uint32_t len)
{
    u_char                  *p;
    ngx_http_wasm_module_t  *hwm;

    hwm = host->hw_mod;

    if (hwm->phase_hooks[phase].len) {
        ngx_log_error(NGX_LOG_WARN, host->cycle->log, 0,
                      "%V host: module \"%V\" redefined \"%s\" phase handler",
                      &host->wh.name, &host->hw_mod->name, phase_name);
    }

    p = ngx_pnalloc(host->cycle->pool, len + 1);
    if (p == NULL) {
        return NGX_ERROR;
    }

    ngx_memcpy(p, buf, len);
    p[len] = 0;

    hwm->phase_hooks[phase].data = p;
    hwm->phase_hooks[phase].len = len;

    ngx_log_debug4(NGX_LOG_DEBUG_HTTP, host->cycle->log, 0,
                   "%V host: add \"%s\" phase handler:"
                   " mod \"%V\" \"%V\"",
                   &host->wh.name, phase_name, &host->hw_mod->name,
                   &hwm->phase_hooks[phase]);

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_http_host_install_request_body_filter(ngx_http_wasm_host_conf_t *host,
    uint8_t *buf, uint32_t len)
{
    u_char                  *p;
    ngx_http_wasm_module_t  *hwm;

    hwm = host->hw_mod;

    if (hwm->request_body_filter.len) {
        ngx_log_error(NGX_LOG_WARN, host->cycle->log, 0,
                      "%V host: module \"%V\" redefined request body filter",
                      &host->wh.name, &host->hw_mod->name);
    }

    p = ngx_pnalloc(host->cycle->pool, len + 1);
    if (p == NULL) {
        return NGX_ERROR;
    }

    ngx_memcpy(p, buf, len);
    p[len] = 0;

    hwm->request_body_filter.data = p;
    hwm->request_body_filter.len = len;

    ngx_log_debug3(NGX_LOG_DEBUG_HTTP, host->cycle->log, 0,
                   "%V host: add request body filter mod \"%V\" \"%V\"",
                   &host->wh.name, &host->hw_mod->name,
                   &hwm->request_body_filter);

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_http_host_install_response_header_filter(ngx_http_wasm_host_conf_t *host,
    uint8_t *buf, uint32_t len)
{
    u_char                  *p;
    ngx_http_wasm_module_t  *hwm;

    hwm = host->hw_mod;

    if (hwm->header_filter.len) {
        ngx_log_error(NGX_LOG_WARN, host->cycle->log, 0,
                      "%V host: module \"%V\" redefined header filter",
                      &host->wh.name, &host->hw_mod->name);
    }

    p = ngx_pnalloc(host->cycle->pool, len + 1);
    if (p == NULL) {
        return NGX_ERROR;
    }

    ngx_memcpy(p, buf, len);
    p[len] = 0;

    hwm->header_filter.data = p;
    hwm->header_filter.len = len;

    ngx_log_debug3(NGX_LOG_DEBUG_HTTP, host->cycle->log, 0,
                   "%V host: add header filter mod \"%V\" \"%V\"",
                   &host->wh.name, &host->hw_mod->name, &hwm->header_filter);

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_http_host_install_response_body_filter(ngx_http_wasm_host_conf_t *host,
    uint8_t *buf, uint32_t len)
{
    u_char                  *p;
    ngx_http_wasm_module_t  *hwm;

    hwm = host->hw_mod;

    if (hwm->response_body_filter.len) {
        ngx_log_error(NGX_LOG_WARN, host->cycle->log, 0,
                      "%V host: module \"%V\" redefined response body filter",
                      &host->wh.name, &host->hw_mod->name);
    }

    p = ngx_pnalloc(host->cycle->pool, len + 1);
    if (p == NULL) {
        return NGX_ERROR;
    }

    ngx_memcpy(p, buf, len);
    p[len] = 0;

    hwm->response_body_filter.data = p;
    hwm->response_body_filter.len = len;

    ngx_log_debug3(NGX_LOG_DEBUG_HTTP, host->cycle->log, 0,
                   "%V host: add response body filter mod \"%V\" \"%V\"",
                   &host->wh.name, &host->hw_mod->name,
                   &hwm->response_body_filter);

    return NGX_OK;
}


static int32_t
ngx_wasm_http_host_set_conf(ngx_wasm_handle_t *self, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    const char                 *phase_name;
    ngx_uint_t                  phase;
    ngx_http_wasm_host_conf_t  *host;

    if (len == 0) {
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V host: set_conf empty hook name", &self->host->name);
        return NGX_ERROR;
    }

    host = (ngx_http_wasm_host_conf_t *) self->host;

    switch (propid) {

    case NGX_WASI_HTTP_MODULE_HOOK_POST_READ:
        phase = NGX_HTTP_POST_READ_PHASE;
        phase_name = "post_read";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_SERVER_REWRITE:
        phase = NGX_HTTP_SERVER_REWRITE_PHASE;
        phase_name = "server_rewrite";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_LOCATION_REWRITE:
        phase = NGX_HTTP_REWRITE_PHASE;
        phase_name = "rewrite";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_PREACCESS:
        phase = NGX_HTTP_PREACCESS_PHASE;
        phase_name = "preaccess";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_ACCESS:
        phase = NGX_HTTP_ACCESS_PHASE;
        phase_name = "access";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_PRECONTENT:
        phase = NGX_HTTP_PRECONTENT_PHASE;
        phase_name = "precontent";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_CONTENT:
        phase = NGX_HTTP_CONTENT_PHASE;
        phase_name = "content";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_LOG:
        phase = NGX_HTTP_LOG_PHASE;
        phase_name = "log";
        break;

    case NGX_WASI_HTTP_MODULE_HOOK_REQUEST_BODY_FILTER:
        return ngx_wasm_http_host_install_request_body_filter(host, buf, len);

    case NGX_WASI_HTTP_MODULE_HOOK_RESPONSE_HEADER_FILTER:
        return ngx_wasm_http_host_install_response_header_filter(host, buf, len);

    case NGX_WASI_HTTP_MODULE_HOOK_RESPONSE_BODY_FILTER:
        return ngx_wasm_http_host_install_response_body_filter(host, buf, len);

    default:
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V host: set conf unknown property: %ui",
                      &self->host->name, propid);
        return NGX_ERROR;
   }

    return ngx_wasm_http_host_install_phase_hook(host, phase, phase_name,
                                                 buf, len);
}


static ngx_int_t
ngx_http_wasm_host_call_module(ngx_http_request_t *r,
    ngx_http_wasm_module_state_t *ctx, ngx_str_t *handler)
{
    ngx_int_t              rc;
    ngx_wasm_val_t         args[2];
    ngx_wasm_call_t        call;
    ngx_http_wasm_host_t  *host;

    host = ctx->host;

    ngx_memzero(&call, sizeof(ngx_wasm_call_t));

    call.host = &host->wh;
    call.name = *handler;
    call.nargs = 2;
    call.args = args;

    call.rc.kind = NGX_WT_I32;

    args[0].kind = NGX_WT_I32;
    args[0].of.i32 = ctx->ce->fd;

    args[1].kind = NGX_WT_I32;
    args[1].of.i32 = ctx->host_fd;

    rc = ngx_wasm_enter(ctx->wi, &call);
    if (rc != NGX_OK) {
        /* some error during wasm execution */
        return NGX_ERROR;
    }

    /* call.rc contains return code from wasm request handler */
    rc = call.rc.of.i32;

    return rc;
}


static ngx_http_wasm_host_rctx_t *
ngx_http_wasm_host_get_ctx(ngx_http_request_t *r)
{
    ngx_http_wasm_host_rctx_t       *rctx;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    rctx = (ngx_http_wasm_host_rctx_t *) ngx_http_get_module_ctx(r,
                                                    ngx_http_wasm_host_module);

    if (rctx) {
        return rctx;
    }

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    rctx = ngx_pcalloc(r->pool, sizeof(ngx_http_wasm_host_rctx_t));
    if (rctx == NULL) {
        return NULL;
    }

    rctx->http_wasm_modules = ngx_pcalloc(r->pool,
                 sizeof(ngx_http_wasm_module_state_t *) * hmcf->modules.nelts);

    ngx_http_set_ctx(r, rctx, ngx_http_wasm_host_module);

    rctx->http_state.request = r;

    return rctx;
}


static void
ngx_http_wasm_host_wasm_module_cleanup(void *data)
{
    ngx_http_wasm_module_state_t  *ctx = data;

    if (ctx->wi == NULL) {
        return;
    }

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, ctx->wi->log, 0,
                   "%V host: deleting instance ptr:%p",
                   &ctx->host->wh.name, ctx->wi);

    ngx_wasm_delete_instance(ctx->wi);

}


static ngx_http_wasm_module_state_t *
ngx_http_wasm_host_get_module_ctx(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod)
{
    ngx_pool_cleanup_t            *cln;
    ngx_http_wasm_module_state_t  *ctx;

    ctx = rctx->http_wasm_modules[hw_mod->own_index];

    if (ctx) {
        return ctx;
    }

    cln = ngx_pool_cleanup_add(r->pool, sizeof(ngx_http_wasm_module_state_t));
    if (cln == NULL) {
        return NULL;
    }

    cln->handler = ngx_http_wasm_host_wasm_module_cleanup;

    ctx = cln->data;
    ngx_memzero(ctx, sizeof(ngx_http_wasm_module_state_t));

    ctx->hw = hw_mod;

    if (ngx_http_wasm_host_init_instance(r, rctx, ctx) != NGX_OK) {
        return NULL;
    }

    rctx->http_wasm_modules[hw_mod->own_index] = ctx;

    return ctx;
}


static ngx_int_t
ngx_http_wasm_host_phase_handler(ngx_http_request_t *r, ngx_uint_t phase,
    const char *phase_name)
{
    ngx_int_t                        rc;
    ngx_uint_t                       i, *indexes;
    ngx_http_wasm_module_t          *hw_modules, *hw_mod;
    ngx_http_wasm_host_rctx_t       *rctx;
    ngx_http_wasm_host_loc_conf_t   *hwlcf;
    ngx_http_wasm_host_main_conf_t  *hmcf;
    ngx_array_t                     *phase_handlers;

    hwlcf = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (hwlcf->phase_handlers == NULL) {
        return NGX_DECLINED;
    }

    phase_handlers = (*hwlcf->phase_handlers)[phase];

    if (phase_handlers == NULL || phase_handlers->nelts == 0) {
        return NGX_DECLINED;
    }

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "wasm_http host: processing phase \"%s\"", phase_name);

    rctx = ngx_http_wasm_host_get_ctx(r);
    if (rctx == NULL) {
        return NGX_ERROR;
    }

    rctx->curr_phase = phase;

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    hw_modules = hmcf->modules.elts;
    indexes = phase_handlers->elts;

    for (i = rctx->next_phase_module; i < phase_handlers->nelts; i++) {

        rctx->next_phase_module = i;

        hw_mod = &hw_modules[indexes[i]];

        rctx->curr_module = ngx_http_wasm_host_get_module_ctx(r, rctx, hw_mod);

        if (hw_mod->phase_hooks[phase].len == 0) {
            /* this modules does not have handler for this phase */
            continue;
        }

        ngx_log_debug2(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "wasm_http host: handling phase \"%s\" "
                       "with \"%V\" module",
                       phase_name, &hw_mod->name);

        rc = ngx_http_wasm_module_phase_handler(r, rctx, hw_mod, phase);

        ngx_log_debug3(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "wasm_http host: module \"%V\" phase \"%s\""
                       " handler rc:%i",
                       &hw_mod->name, phase_name, rc);

        if (phase == NGX_HTTP_CONTENT_PHASE) {
            /* only one module can handle content phase */
            goto done;
        }

        if (rc == NGX_DECLINED) {
            /* proceed to the next handler of the current phase. */
            continue;
        }

        if (rc == NGX_OK) {
            /* proceed to the next phase */
            goto done;
        }

        if (rc == NGX_AGAIN || rc == NGX_DONE) {
            /* suspend phase handling till some event in future */

            ngx_log_debug2(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                           "wasm_http host: phase \"%s\" suspended rc:%i",
                           phase_name, rc);
            return rc;
        }

        /* request finalization code */
        goto done;
    }


    rc = NGX_DECLINED;

done:

    /* start next phase with first module */
    rctx->next_phase_module = 0;

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "wasm_http host: phase \"%s\" completed rc:%i",
                   phase_name, rc);

    return rc;
}


static ngx_int_t
ngx_http_wasm_module_phase_handler(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod,
    ngx_uint_t phase)
{
    ngx_int_t                      rc;
    ngx_http_wasm_module_state_t  *ctx;

    /* this phase handler is common for all wasm modules using this host */

    ctx = ngx_http_wasm_host_get_module_ctx(r, rctx, hw_mod);
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    rc = ngx_http_wasm_host_call_module(r, ctx, &hw_mod->phase_hooks[phase]);

    if (rctx->http_state.reject_msg.len) {
        return ngx_http_wasm_host_reject_request(r,
                                                 &rctx->http_state.reject_msg);
    }

    return rc;
}


static ngx_int_t
ngx_http_wasm_post_read(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_POST_READ_PHASE,
                                            "post_read");
}


static ngx_int_t
ngx_http_wasm_server_rewrite(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_SERVER_REWRITE_PHASE,
                                            "server_rewrite");
}


static ngx_int_t
ngx_http_wasm_rewrite(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_REWRITE_PHASE,
                                            "rewrite");
}


static ngx_int_t
ngx_http_wasm_preaccess(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_PREACCESS_PHASE,
                                            "preaccess");
}


static ngx_int_t
ngx_http_wasm_access(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_ACCESS_PHASE, "access");
}


static ngx_int_t
ngx_http_wasm_precontent(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_PRECONTENT_PHASE,
                                            "precontent");
}


static ngx_int_t
ngx_http_wasm_content(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_CONTENT_PHASE,
                                            "content");
}


static ngx_int_t
ngx_http_wasm_log(ngx_http_request_t *r)
{
    return ngx_http_wasm_host_phase_handler(r, NGX_HTTP_LOG_PHASE, "log");
}


static ngx_int_t
ngx_http_wasm_host_postconf(ngx_conf_t *cf)
{
    ngx_uint_t                  i;
    ngx_http_handler_pt        *h;
    ngx_http_core_main_conf_t  *cmcf;

    cmcf = ngx_http_conf_get_module_main_conf(cf, ngx_http_core_module);

    struct {
        ngx_uint_t             phase;
        ngx_http_handler_pt    handler;
    } bindings[] = {
        { NGX_HTTP_POST_READ_PHASE, ngx_http_wasm_post_read },
        { NGX_HTTP_SERVER_REWRITE_PHASE, ngx_http_wasm_server_rewrite },
        { NGX_HTTP_REWRITE_PHASE, ngx_http_wasm_rewrite },
        { NGX_HTTP_PREACCESS_PHASE, ngx_http_wasm_preaccess },
        { NGX_HTTP_ACCESS_PHASE, ngx_http_wasm_access },
        { NGX_HTTP_PRECONTENT_PHASE, ngx_http_wasm_precontent },
        { NGX_HTTP_CONTENT_PHASE, ngx_http_wasm_content },
        { NGX_HTTP_LOG_PHASE, ngx_http_wasm_log },
    };

    for (i = 0; i < sizeof(bindings)/sizeof(bindings[0]); i++) {
        h = ngx_array_push(&cmcf->phases[bindings[i].phase].handlers);
        if (h == NULL) {
            return NGX_ERROR;
        }

        *h = bindings[i].handler;
    }

    ngx_http_next_header_filter = ngx_http_top_header_filter;
    ngx_http_top_header_filter = ngx_http_wasm_host_header_filter;

    ngx_http_next_body_filter = ngx_http_top_body_filter;
    ngx_http_top_body_filter = ngx_http_wasm_host_response_body_filter;

    ngx_http_next_request_body_filter = ngx_http_top_request_body_filter;
    ngx_http_top_request_body_filter = ngx_http_wasm_host_request_body_filter;

    return NGX_OK;
}


static void
ngx_http_wasm_host_conf_cleanup(void *data)
{
    ngx_wasm_instance_t *wi = data;

    if (wi == NULL) {
        return;
    }

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, wi->log, 0,
                   "wasm_http_conf host: deleting instance ptr:%p", wi);

    ngx_wasm_delete_instance(wi);
}


static ngx_int_t
ngx_http_wasm_host_init_worker(ngx_cycle_t *cycle)
{
    ngx_int_t                        rc;
    ngx_uint_t                       i;
    ngx_pool_t                      *pool;
    ngx_wasm_conf_t                 *wcf;
    ngx_wasm_call_t                  call;
    ngx_wasm_module_t               *wasm_modules, *wm;
    ngx_pool_cleanup_t              *cln;
    ngx_wasm_instance_t             *wi;
    ngx_http_wasm_module_t          *hw_modules, *hw_mod;
    ngx_http_wasm_host_conf_t       *whcf;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    ngx_wasm_val_t                   args[1];

    hmcf = ngx_http_cycle_get_module_main_conf(ngx_cycle,
                                               ngx_http_wasm_host_module);

    hw_modules = hmcf->modules.elts;

    if (hmcf->modules.nelts == 0) {
        ngx_log_debug0(NGX_LOG_DEBUG_CORE, cycle->log, 0,
                       "wasm_http: no http wasm modules found, done");
        return NGX_OK;
    }

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(cycle->conf_ctx,
                                                ngx_wasm_core_module);

    wasm_modules = wcf->modules.elts;

    pool = ngx_create_pool(NGX_DEFAULT_POOL_SIZE, cycle->log);
    if (pool == NULL) {
        return NGX_ERROR;
    }

    ngx_log_debug1(NGX_LOG_DEBUG_CORE, cycle->log, 0,
                   "wasm_http: init wasm modules: %ui total",
                   hmcf->modules.nelts);

    for (i = 0; i < hmcf->modules.nelts; i++) {

        hw_mod = &hw_modules[i];

        wm = &wasm_modules[hw_mod->index];

        if (wm->init_func.len == 0) {
            ngx_log_debug1(NGX_LOG_DEBUG_CORE, cycle->log, 0,
                           "wasm_http: http wasm module \"%V\""
                           " missing init function, skipped",
                           &hw_mod->name);
            continue;
        }

        ngx_log_debug1(NGX_LOG_DEBUG_CORE, cycle->log, 0,
                       "wasm_http: initializing http wasm module \"%V\"",
                       &hw_mod->name);

        cln = ngx_pool_cleanup_add(pool, 0);
        if (cln == NULL) {
            rc = NGX_ERROR;
            goto done;
        }

        cln->handler = ngx_http_wasm_host_conf_cleanup;
        cln->data = NULL;

        wi = ngx_wasm_create_instance(wm, pool, cycle->log);
        if (wi == NULL) {
            rc = NGX_ERROR;
            goto done;
        }

        cln->data = wi;

        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, wi->log, 0,
                   "wasm_htp_conf host: new instance created ptr:%p", wi);

        whcf = ngx_http_wasm_conf_host_create(wi, "wasm_http_conf");
        if (whcf == NULL) {
            rc = NGX_ERROR;
            goto done;
        }

        whcf->cycle = cycle;
        whcf->hw_mod = hw_mod;

        ngx_memzero(&call, sizeof(ngx_wasm_call_t));

        call.host = &whcf->wh;
        call.name = wm->init_func;
        call.nargs = 1;
        call.args = args;
        call.rc.kind = NGX_WT_I32;

        args[0].kind = NGX_WT_I32;
        args[0].of.i32 = whcf->conf_fd;

        rc = ngx_wasm_enter(wi, &call);
        if (rc == NGX_ERROR) {
            rc = NGX_ERROR;
            goto done;
        }
    }

    rc = NGX_OK;

done:

    ngx_destroy_pool(pool);

    return rc;
}


static void
ngx_wasm_host_handle_events(ngx_event_t *ev)
{
    ngx_log_debug0(NGX_LOG_DEBUG_EVENT, ev->log, 0,
                   "wasm http host: event handler");

    ngx_http_core_run_phases(ev->data);
}


static ngx_int_t
ngx_http_wasm_host_init_instance(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_state_t *ctx)
{
    int32_t             fd;
    ngx_wasm_conf_t    *wcf;
    ngx_wasm_module_t  *wasm_modules, *wm;

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(ngx_cycle->conf_ctx,
                                                ngx_wasm_core_module);
    wasm_modules = wcf->modules.elts;

    wm = &wasm_modules[ctx->hw->index];

    ctx->wi = ngx_wasm_create_instance(wm, r->pool, r->connection->log);
    if (ctx->wi == NULL) {
        return NGX_ERROR;
    }

    /* each wasm http module has own host object with own descriptor set */
    ctx->host = ngx_http_wasm_host_create(ctx->wi, r, "wasm_http");
    if (ctx->host == NULL) {
        return NGX_ERROR;
    }

    ngx_log_debug2(NGX_LOG_DEBUG_HTTP, ctx->wi->log, 0,
                   "%V host: new instance created ptr:%p", &ctx->host->wh.name,
                   ctx->wi);

    ctx->ce = ngx_pcalloc(r->pool, sizeof(ngx_wasm_call_env_t));
    if (ctx->ce == NULL) {
        return NGX_ERROR;
    }

    ctx->ce->state = &ctx->wstate;

    ctx->ce->fd = ngx_wasm_host_open_object(&ctx->host->wh,
                                            "ngx::call_env::void", ctx->ce);
    if (ctx->ce->fd == -1) {
        ngx_log_error(NGX_LOG_EMERG, r->connection->log, 0,
                      "wasm filter failed to open call void env on host");
        return NGX_ERROR;
    }

    /* TODO: pass open fd to wasm instead of open() inside ? */
    fd = ngx_wasm_host_create_api_context(&ctx->host->wh,
                                          "ngx::http::request",
                                          &rctx->http_state);
    if (fd == -1) {
        return NGX_ERROR;
    }

    ctx->host_fd = fd;

    ctx->ev.handler = ngx_wasm_host_handle_events;
    ctx->ev.log = r->connection->log;
    ctx->ev.data = r;

    ctx->host->wh.fd_event = &ctx->ev;

    return NGX_OK;
}


static void
ngx_http_wasm_host_body_handler(ngx_http_request_t *r)
{
    ngx_int_t                  rc;
    ngx_str_t                 *func;
    ngx_wasm_http_ctx_t       *hs;
    ngx_http_wasm_host_rctx_t  *rctx;

    rctx = (ngx_http_wasm_host_rctx_t *) ngx_http_get_module_ctx(r,
                                                    ngx_http_wasm_host_module);
    hs = &rctx->http_state;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "wasm_http host: request body handler isc:%i",
                   hs->in_syscall);

    hs->post_body_called = 1;

    if (r->request_body == NULL) {
        ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                      "wasm_http host: request body handler empty body");
        ngx_http_finalize_request(r, NGX_HTTP_INTERNAL_SERVER_ERROR);
        return;
    }

    if (!hs->in_syscall) {
        /* called from event loop after the body was finally read */

        func = (ngx_str_t *) hs->body_handler_params;
        rc = ngx_http_wasm_host_call_module(r, rctx->curr_module, func);

    } else {
        rc = NGX_OK;
    }

    ngx_http_finalize_request(r, rc);
}


static ngx_int_t
ngx_http_wasm_response_body_filter_module(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod,
    ngx_chain_t *in)
{
    int32_t                        obj_fd;
    int64_t                        res;
    ngx_int_t                      rc;
    ngx_wasm_val_t                 args[3];
    ngx_wasm_call_t                call;
    ngx_http_wasm_host_t          *host;
    ngx_wasm_filter_env_t          fenv;
    ngx_http_wasm_module_state_t  *ctx;

    ctx = ngx_http_wasm_host_get_module_ctx(r, rctx, hw_mod);
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    host = ctx->host;

    fenv.in = in;
    fenv.freep = &rctx->free;
    fenv.outp = &rctx->out;
    fenv.tag = (ngx_buf_tag_t) &ngx_http_wasm_host_module;

    obj_fd = ngx_wasm_host_create_api_context(&host->wh, "ngx::filter",
                                              &fenv);
    if (obj_fd == -1) {
        return NGX_ERROR;
    }

    ngx_memzero(&call, sizeof(ngx_wasm_call_t));

    call.host = &host->wh;
    call.name = hw_mod->response_body_filter;
    call.nargs = 3;
    call.args = args;

    call.rc.kind = NGX_WT_I64;

    args[0].kind = NGX_WT_I32;
    args[0].of.i32 = ctx->ce->fd;

    args[1].kind = NGX_WT_I32;
    args[1].of.i32 = ctx->host_fd;

    args[2].kind = NGX_WT_I32;
    args[2].of.i32 = obj_fd;

    rc = ngx_wasm_enter(ctx->wi, &call);

    (void) ngx_wasm_host_delete_object(&host->wh, obj_fd);

    if (rc != NGX_OK) {
        return NGX_ERROR;
    }

    res = call.rc.of.i64;
    if (res < 0) {
        ngx_log_error(NGX_LOG_EMERG, r->connection->log, 0,
                      "wasm_http host: response body filter failed rc:%L",
                      res);
        return NGX_ERROR;
    }

    return NGX_OK;
}


/* TODO: same as response body filter ? */
static ngx_int_t
ngx_http_wasm_request_body_filter_module(ngx_http_request_t *r,
    ngx_http_wasm_host_rctx_t *rctx, ngx_http_wasm_module_t *hw_mod,
    ngx_chain_t *in)
{
    int32_t                        obj_fd;
    int64_t                        res;
    ngx_int_t                      rc;
    ngx_wasm_val_t                 args[3];
    ngx_wasm_call_t                call;
    ngx_http_wasm_host_t          *host;
    ngx_wasm_filter_env_t          fenv;
    ngx_http_wasm_module_state_t  *ctx;

    ctx = ngx_http_wasm_host_get_module_ctx(r, rctx, hw_mod);
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    host = ctx->host;

    fenv.in = in;
    fenv.freep = &rctx->free;
    fenv.outp = &rctx->rb_out;
    fenv.tag = (ngx_buf_tag_t) &ngx_http_wasm_host_module;

    obj_fd = ngx_wasm_host_create_api_context(&host->wh, "ngx::filter",
                                              &fenv);
    if (obj_fd == -1) {
        return NGX_ERROR;
    }

    ngx_memzero(&call, sizeof(ngx_wasm_call_t));

    call.host = &host->wh;
    call.name = hw_mod->request_body_filter;
    call.nargs = 3;
    call.args = args;

    call.rc.kind = NGX_WT_I64;

    args[0].kind = NGX_WT_I32;
    args[0].of.i32 = ctx->ce->fd;

    args[1].kind = NGX_WT_I32;
    args[1].of.i32 = ctx->host_fd;

    args[2].kind = NGX_WT_I32;
    args[2].of.i32 = obj_fd;

    rc = ngx_wasm_enter(ctx->wi, &call);

    (void) ngx_wasm_host_delete_object(&host->wh, obj_fd);

    if (rc != NGX_OK) {
        return NGX_ERROR;
    }

    res = call.rc.of.i64;
    if (res < 0) {
        ngx_log_error(NGX_LOG_EMERG, r->connection->log, 0,
                      "wasm_http host: request body filter failed rc:%L",
                      res);
        return NGX_ERROR;
    }

    return NGX_OK;
}


/* ngx_chain_add_copy(), but takes free bufs from the 'freep' chain */
static ngx_int_t
ngx_chain_add_copy_from(ngx_pool_t *pool, ngx_chain_t **chain, ngx_chain_t *in,
    ngx_chain_t **freep)
{
    ngx_chain_t  *cl, **ll;

    ll = chain;

    for (cl = *chain; cl; cl = cl->next) {
        ll = &cl->next;
    }

    while (in) {
        cl = ngx_chain_get_free_buf(pool, freep);
        if (cl == NULL) {
            *ll = NULL;
            return NGX_ERROR;
        }

        cl->buf = in->buf;
        *ll = cl;
        ll = &cl->next;
        in = in->next;
    }

    *ll = NULL;

    return NGX_OK;
}


static ngx_int_t
ngx_http_wasm_host_request_body_filter(ngx_http_request_t *r, ngx_chain_t *in)
{
    ngx_int_t                        rc;
    ngx_uint_t                       i, *indexes;
    ngx_chain_t                     *chain;
    ngx_http_wasm_module_t          *hw_modules, *hw_mod;
    ngx_http_wasm_host_rctx_t       *rctx;
    ngx_http_wasm_host_loc_conf_t   *hwlcf;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    hwlcf = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (hwlcf->request_body_filters == NULL
        || hwlcf->request_body_filters->nelts == 0)
    {
        return ngx_http_next_request_body_filter(r, in);
    }

    rctx = ngx_http_wasm_host_get_ctx(r);
    if (rctx == NULL) {
        return NGX_ERROR;
    }

    if (rctx->http_state.in_syscall) {
        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                       "wasm_http host: request body filter nested host call");

        /* we are going to dealy data buffers a bit... */
        r->request_body->filter_need_buffering = 1;

        if (ngx_chain_add_copy_from(r->pool, &rctx->in, in, &rctx->free)
            != NGX_OK)
        {
            return NGX_HTTP_INTERNAL_SERVER_ERROR;
        }

        ngx_post_event(r->connection->read, &ngx_posted_events);

        return ngx_http_next_request_body_filter(r, NULL);
    }

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    hw_modules = hmcf->modules.elts;
    indexes = hwlcf->request_body_filters->elts;

    chain = rctx->in ? rctx->in : in;

    for (i = rctx->next_request_body_filter_module;
         i < hwlcf->request_body_filters->nelts;
         i++)
    {
        hw_mod = &hw_modules[indexes[i]];
        rctx->next_request_body_filter_module = i;

        if (hw_mod->request_body_filter.len == 0) {
            ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                           "http wasm module \"%V\" did not install request "
                           "body filter used in configuration", &hw_mod->name);

            continue;
        }

        rc = ngx_http_wasm_request_body_filter_module(r, rctx, hw_mod, chain);
        if (rc != NGX_OK) {
            return rc;
        }

        chain = rctx->rb_out;
    }

    rc = ngx_http_next_request_body_filter(r, rctx->rb_out);

    ngx_chain_update_chains(r->pool, &rctx->free, &rctx->busy, &rctx->rb_out,
                            (ngx_buf_tag_t) &ngx_http_wasm_host_module);

    return rc;
}


static ngx_int_t
ngx_http_wasm_host_header_filter(ngx_http_request_t *r)
{
    ngx_int_t                        rc;
    ngx_uint_t                       i, *indexes;
    ngx_http_wasm_module_t          *hw_modules, *hw_mod;
    ngx_http_wasm_host_rctx_t       *rctx;
    ngx_http_wasm_module_state_t    *ctx;
    ngx_http_wasm_host_loc_conf_t   *hwlcf;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    hwlcf = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (hwlcf->header_filters == NULL || hwlcf->header_filters->nelts == 0) {
        return ngx_http_next_header_filter(r);
    }

    rctx = ngx_http_wasm_host_get_ctx(r);
    if (rctx == NULL) {
        return NGX_ERROR;
    }

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    hw_modules = hmcf->modules.elts;
    indexes = hwlcf->header_filters->elts;

    for (i = rctx->next_header_filter_module;
         i < hwlcf->header_filters->nelts;
         i++)
    {
        rctx->next_header_filter_module = i;

        hw_mod = &hw_modules[indexes[i]];

        if (hw_mod->header_filter.len == 0) {
            ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                           "http wasm module \"%V\" did not install "
                           "header filter used in configuration",
                           &hw_mod->name);
            continue;
        }

        ctx = ngx_http_wasm_host_get_module_ctx(r, rctx, hw_mod);
        if (ctx == NULL) {
            return NGX_ERROR;
        }

        rc = ngx_http_wasm_host_call_module(r, ctx, &hw_mod->header_filter);
        if (rc != NGX_OK) {
            return rc;
        }
    }

    return ngx_http_next_header_filter(r);
}


static ngx_int_t
ngx_http_wasm_host_response_body_filter(ngx_http_request_t *r, ngx_chain_t *in)
{
    ngx_int_t                        rc;
    ngx_uint_t                       i, *indexes;
    ngx_chain_t                     *chain;
    ngx_http_wasm_module_t          *hw_modules, *hw_mod;
    ngx_http_wasm_host_rctx_t       *rctx;
    ngx_http_wasm_host_loc_conf_t   *hwlcf;
    ngx_http_wasm_host_main_conf_t  *hmcf;

    if (in == NULL || r->header_only) {
        return ngx_http_next_body_filter(r, in);
    }

    hwlcf = ngx_http_get_module_loc_conf(r, ngx_http_wasm_host_module);

    if (hwlcf->response_body_filters == NULL
        || hwlcf->response_body_filters->nelts == 0)
    {
        return ngx_http_next_body_filter(r, in);
    }

    rctx = ngx_http_wasm_host_get_ctx(r);
    if (rctx == NULL) {
        return NGX_ERROR;
    }

    hmcf = ngx_http_get_module_main_conf(r, ngx_http_wasm_host_module);

    hw_modules = hmcf->modules.elts;
    indexes = hwlcf->response_body_filters->elts;

    chain = in;

    for (i = rctx->next_response_body_filter_module;
         i < hwlcf->response_body_filters->nelts;
         i++)
    {
        hw_mod = &hw_modules[indexes[i]];
        rctx->next_response_body_filter_module = i;

        if (hw_mod->response_body_filter.len == 0) {
            ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                          "http wasm module \"%V\" did not install response "
                          "body filter used in configuration", &hw_mod->name);

            continue;
        }

        rc = ngx_http_wasm_response_body_filter_module(r, rctx, hw_mod, chain);
        if (rc != NGX_OK) {
            return rc;
        }

        chain = rctx->out;
    }

    rc = ngx_http_next_body_filter(r, rctx->out);

    ngx_chain_update_chains(r->pool, &rctx->free, &rctx->busy, &rctx->out,
                            (ngx_buf_tag_t) &ngx_http_wasm_host_module);

    return rc;
}


static char *
ngx_http_wasm_host_variable(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    char                               *rv;
    size_t                             n;
    ngx_str_t                          name,  *value;
    ngx_uint_t                         i;
    ngx_http_variable_t               *var;
    ngx_http_complex_value_t          *cv;
    ngx_http_wasm_host_var_ctx_t      *ctx;
    ngx_http_compile_complex_value_t   ccv;

    value = cf->args->elts;

    if (cf->args->nelts < 4) {
        return "invalid number of arguments";
    }

    ctx = ngx_pcalloc(cf->pool, sizeof(ngx_http_wasm_host_var_ctx_t));
    if (ctx == NULL) {
        return NGX_CONF_ERROR;
    }

    rv = ngx_conf_set_wasm_ref_slot(cf, cmd, &ctx->fc);

    if (rv != NGX_CONF_OK) {
        return rv;
    }

    name = value[3];

    if (name.data[0] != '$') {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "invalid variable name \"%V\"", &name);
        return NGX_CONF_ERROR;
    }

    name.len--;
    name.data++;

    var = ngx_http_add_variable(cf, &name, NGX_HTTP_VAR_CHANGEABLE);
    if (var == NULL) {
        return NGX_CONF_ERROR;
    }

    n = cf->args->nelts - 3;

    if (ngx_array_init(&ctx->arg_vars, cf->pool, n,
                       sizeof(ngx_http_complex_value_t))
        != NGX_OK)
    {
        return NGX_CONF_ERROR;
    }

    for (i = 4; i < cf->args->nelts; i++) {

        cv = ngx_array_push(&ctx->arg_vars);
        if (cv == NULL) {
            return NGX_CONF_ERROR;
        }

        ngx_memzero(&ccv, sizeof(ngx_http_compile_complex_value_t));

        ccv.cf = cf;
        ccv.value = &value[i];
        ccv.complex_value = cv;

        if (ngx_http_compile_complex_value(&ccv) != NGX_OK) {
            return NGX_CONF_ERROR;
        }
    }

    var->get_handler = ngx_http_wasm_host_variable_handler;
    var->data = (uintptr_t) ctx;

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_http_wasm_host_variable_handler(ngx_http_request_t *r,
    ngx_http_variable_value_t *v, uintptr_t data)
{
    ngx_http_wasm_host_var_ctx_t *ctx = (ngx_http_wasm_host_var_ctx_t *) data;

    int32_t                fda;
    ngx_int_t              rc;
    ngx_str_t              result;
    ngx_wasm_val_t         args[1];
    ngx_wasm_call_t        call;
    ngx_wasm_call_env_t   *ce;
    ngx_wasm_instance_t   *wi;
    ngx_http_wasm_host_t  *host;

    ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                   "wasm vars variable handler nvars:%ui",
                   ctx->arg_vars.nelts);

    wi = ngx_http_wasm_host_variable_get_instance(r, ctx->fc.wm);
    if (wi == NULL) {
        return NGX_ERROR;
    }

    /* setup host environment */

    host = ngx_http_wasm_host_create(wi, r, "wasm vars");
    if (host == NULL) {
        return NGX_ERROR;
    }

    ce = ngx_http_wasm_host_variable_create_call_env(r, ctx, &result);
    if (ce == NULL) {
        return NGX_ERROR;
    }

    fda = ngx_wasm_host_open_object(&host->wh, "ngx::call_env::argv", ce);
    if (fda == -1) {
        ngx_log_error(NGX_LOG_EMERG, r->connection->log, 0,
                      "wasm vars failed to open call env argv on host");
        return NGX_ERROR;
    }

    /* call the function that computes the variable */

    ngx_memzero(&call, sizeof(ngx_wasm_call_t));

    call.host = &host->wh;
    call.name = ctx->fc.func; /* as specified in configuration file */

    call.args = args;
    call.nargs = 1;

    call.rc.kind = NGX_WT_I32;

    /* func(args_fd); */

    args[0].kind = NGX_WT_I32;
    args[0].of.i32 = fda;

    rc = ngx_wasm_enter(wi, &call);

    if (rc != NGX_OK) {
        goto wasm_failed;
    }

    if (call.rc.of.i32 < 0) {
        ngx_log_error(NGX_LOG_NOTICE, r->connection->log, 0,
                      "wasm call failed with rc:%d", call.rc.of.i32);
        goto nores;
    }

    /* the function was succesfully called, obtain results */

    v->len = result.len;
    v->data = ngx_pstrdup(r->pool, &result);
    if (v->data == NULL) {
        goto failed;
    }

    /* cleanup */
    if (ctx->fc.wm->reactor == NULL) {
        ngx_wasm_delete_instance(wi);
    }

    v->valid = 1;
    v->no_cacheable = 0;
    v->not_found = 0;

    return NGX_OK;

nores:

    v->not_found = 1;

    if (ctx->fc.wm->reactor == NULL) {
        ngx_wasm_delete_instance(wi);
    }

    return NGX_OK;

wasm_failed:

    if (rc == NGX_ABORT) {
        /* the wasm instance is heavily broken: trapped or killed by signal */
        if (ctx->fc.wm->reactor) {
            ngx_wasm_delete_instance(ctx->fc.wm->reactor);
            ctx->fc.wm->reactor = NULL;
        }
    }

failed:

    if (ctx->fc.wm->reactor == NULL) {
        ngx_wasm_delete_instance(wi);
    }

    return NGX_ERROR;
}


static ngx_wasm_instance_t *
ngx_http_wasm_host_variable_get_instance(ngx_http_request_t *r,
    ngx_wasm_module_t *wm)
{
    ngx_wasm_instance_t  *wi;

    if (!wm->is_reactor) {
        wi = ngx_wasm_create_instance(wm, r->pool, r->connection->log);
        if (wi == NULL) {
            return NULL;
        }

        return wi;
    }

    if (wm->reactor) {
        return wm->reactor;
    }

    ngx_log_error(NGX_LOG_NOTICE, r->connection->log, 0,
                  "wasm reactor recreating instance");

    /* the instance was destroyed due to fatal error, re-create */
    wi = ngx_wasm_create_instance(wm, ngx_cycle->pool, ngx_cycle->log);
    if (wi == NULL) {
        return NULL;
    }

    wm->reactor = wi;

    return wi;
}


static ngx_wasm_call_env_t *
ngx_http_wasm_host_variable_create_call_env(ngx_http_request_t *r,
    ngx_http_wasm_host_var_ctx_t *ctx, ngx_str_t *result)
{
    ngx_str_t                 *str;
    ngx_uint_t                 i, n;
    ngx_array_t               *argv;
    ngx_wasm_call_env_t       *ce;
    ngx_http_complex_value_t  *vars;

    ce = ngx_pcalloc(r->pool, sizeof(ngx_wasm_call_env_t));
    if (ce == NULL) {
        return NULL;
    }

    ngx_memzero(result, sizeof(ngx_str_t));

    ce->result = result;

    n = ngx_max(1, ctx->arg_vars.nelts);

    argv = ngx_array_create(r->pool, n, sizeof(ngx_str_t));
    if (argv == NULL) {
        return NULL;
    }

    vars = ctx->arg_vars.elts;

    for (i = 0; i < ctx->arg_vars.nelts; i++) {

        str = ngx_array_push(argv);

        if (ngx_http_complex_value(r, &vars[i], str) != NGX_OK) {
            return NULL;
        }
    }

    ce->args = argv;
    ce->state = &ctx->wstate;

    return ce;
}


static void *
ngx_http_wasm_host_create_main_conf(ngx_conf_t *cf)
{
    ngx_uint_t                        i, nelts;
    ngx_wasm_handle_ops_t           **ops;
    ngx_http_wasm_host_main_conf_t   *hmcf;

    hmcf = ngx_pcalloc(cf->pool, sizeof(ngx_http_wasm_host_main_conf_t));
    if (hmcf == NULL) {
        return NULL;
    }

    if (ngx_array_init(&hmcf->modules, cf->pool, 4,
                       sizeof(ngx_http_wasm_module_t))
        != NGX_OK)
    {
        return NULL;
    }

    nelts = sizeof(ngx_http_wasm_apis) / sizeof(ngx_http_wasm_apis[0]);

    if (ngx_array_init(&hmcf->apis, cf->pool, nelts + 5,
                       sizeof(ngx_wasm_handle_ops_t *))
        != NGX_OK)
    {
        return NULL;
    }

    for (i = 0; i < nelts; i++) {
        ops = ngx_array_push(&hmcf->apis);
        *ops = &ngx_http_wasm_apis[i];
    }

    if (ngx_http_wasm_host_add_ops(&hmcf->apis, ngx_wasm_log_ops) != NGX_OK) {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(&hmcf->apis, ngx_wasm_call_env_ops)
        != NGX_OK)
    {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(&hmcf->apis, ngx_wasm_event_ops) != NGX_OK) {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(&hmcf->apis, ngx_wasm_connection_ops)
        != NGX_OK)
    {
        return NULL;
    }

    if (ngx_http_wasm_host_add_ops(&hmcf->apis, ngx_wasm_filter_ops) != NGX_OK)
    {
        return NULL;
    }

    ops = ngx_array_push(&hmcf->apis);
    *ops = ngx_wasm_filter_ops;

    if (ngx_http_wasm_host_load_modules(cf, hmcf) != NGX_CONF_OK) {
        return NGX_CONF_ERROR;
    }

    /* other modules may append more HTTP APIs */
    return hmcf;
}


static char *
ngx_http_wasm_host_load_modules(ngx_conf_t *cf, void *conf)
{
    ngx_http_wasm_host_main_conf_t *hmcf = conf;

    ngx_uint_t              *http_modules_ind, i, idx;
    ngx_wasm_conf_t         *wcf;
    ngx_wasm_module_t       *wasm_modules, *wm;
    ngx_http_wasm_module_t  *hw_mod;

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(cf->cycle->conf_ctx,
                                                ngx_wasm_core_module);
    if (wcf == NULL) {
        /* no wasm_modules{} block at all */
        return NGX_CONF_OK;
    }

    wasm_modules = wcf->modules.elts;
    http_modules_ind = wcf->http_modules.elts;

    /* for each wasm modules in 'http' subsystem, create http wasm module */
    for (i = 0; i < wcf->http_modules.nelts; i++) {
        idx = http_modules_ind[i];
        wm = &wasm_modules[idx];

        hw_mod = ngx_array_push(&hmcf->modules);
        if (hw_mod == NULL) {
            return NGX_CONF_ERROR;
        }

        ngx_memzero(hw_mod, sizeof(ngx_http_wasm_module_t));

        hw_mod->own_index = hw_mod
                        - ((ngx_http_wasm_module_t *) hmcf->modules.elts);

        hw_mod->index = idx;

        hw_mod->name.data = ngx_pstrdup(cf->pool, &wm->ns);
        if (hw_mod->name.data == NULL) {
            return NGX_CONF_ERROR;
        }

        hw_mod->name.len = wm->ns.len;

        ngx_str_set(&wm->init_func, "ngx:wasi/http-module#init-worker");
    }

    return NGX_CONF_OK;
}


static void *
ngx_http_wasm_host_create_loc_conf(ngx_conf_t *cf)
{
    ngx_http_wasm_host_loc_conf_t *conf;

    conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_wasm_host_loc_conf_t));
    if (conf == NULL) {
        return NULL;
    }

    conf->env.env_params = NGX_CONF_UNSET_PTR;
    conf->env.argv = NGX_CONF_UNSET_PTR;

    conf->phase_handlers = NGX_CONF_UNSET_PTR;
    conf->header_filters = NGX_CONF_UNSET_PTR;
    conf->request_body_filters = NGX_CONF_UNSET_PTR;
    conf->response_body_filters = NGX_CONF_UNSET_PTR;

    /*
     * set by ngx_pcalloc():
     *
     */

    return conf;
}


static char *
ngx_http_wasm_host_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
{
    ngx_http_wasm_host_loc_conf_t *prev = parent;
    ngx_http_wasm_host_loc_conf_t *conf = child;

    ngx_conf_merge_ptr_value(conf->env.env_params, prev->env.env_params, NULL);
    ngx_conf_merge_ptr_value(conf->env.argv, prev->env.argv, NULL);

    ngx_conf_merge_ptr_value(conf->phase_handlers, prev->phase_handlers, NULL);

    ngx_conf_merge_ptr_value(conf->header_filters, prev->header_filters, NULL);

    ngx_conf_merge_ptr_value(conf->request_body_filters,
                             prev->request_body_filters, NULL);

    ngx_conf_merge_ptr_value(conf->response_body_filters,
                             prev->response_body_filters, NULL);

    return NGX_CONF_OK;
}
