
/*
 * Copyright (C) 2024 Web Server LLC
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_wasm.h>

#include <ngx_wasm_serial.h>
#include <ngx_wasm_call_env.h>
#include <wasi/ngx_wasi_call_env.h>


// TODO: log all errors

enum {
    NGX_WASM_HOST_CALL_ENV_KV_TAG = 0x43454E56, /* CENV */
    NGX_WASM_HOST_CALL_ENV_ARGV_TAG,
    NGX_WASM_HOST_CALL_ENV_VOID_TAG,
};


typedef struct {
    ngx_kv_array_t        kva;
    ngx_str_array_t       sa;
    uint32_t             *state;
    ngx_str_t            *result;
} ngx_wasm_call_env_ctx_t;


static int32_t ngx_wasm_host_open_kv_env(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_host_open_argv_env(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);
static int32_t ngx_wasm_host_open_void_env(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *ph);

static int32_t ngx_wasm_call_env_set(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_call_env_get(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);

static int32_t ngx_wasm_host_read_kv_env(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);
static int32_t ngx_wasm_host_read_argv_env(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);


ngx_wasm_handle_ops_t ngx_wasm_call_env_ops[] = {

    {   .type = ngx_string("ngx::call_env::kv"),
        .tag = NGX_WASM_HOST_CALL_ENV_KV_TAG,
        .open = ngx_wasm_host_open_kv_env,
        .read = ngx_wasm_host_read_kv_env,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_call_env_get,
        .set = ngx_wasm_call_env_set,
        .close = ngx_wasm_host_close_stub,
    },

    {   .type = ngx_string("ngx::call_env::argv"),
        .tag = NGX_WASM_HOST_CALL_ENV_ARGV_TAG,
        .open = ngx_wasm_host_open_argv_env,
        .read = ngx_wasm_host_read_argv_env,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_call_env_get,
        .set = ngx_wasm_call_env_set,
        .close = ngx_wasm_host_close_stub,
    },

    {   .type = ngx_string("ngx::call_env::void"),
        .tag = NGX_WASM_HOST_CALL_ENV_VOID_TAG,
        .open = ngx_wasm_host_open_void_env,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_call_env_get,
        .set = ngx_wasm_call_env_set,
        .close = ngx_wasm_host_close_stub,
    },

    ngx_wasm_host_null_ops
};


static int32_t
ngx_wasm_host_open_kv_env(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    ngx_wasm_call_env_t      *ea;
    ngx_wasm_call_env_ctx_t  *ctx;

    ea = ngx_wasm_host_get_api_context(self, ph);
    if (ea == NULL) {
        return NGX_ERROR;
    }

    ctx = ngx_pcalloc(self->host->pool, sizeof(ngx_wasm_call_env_ctx_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ctx->result = ea->result;
    ctx->state = ea->state;

    ngx_kv_array_init_array_encoder(&ctx->kva, ea->args->nelts, ea->args->elts);

    self->data = ctx;

    return NGX_OK;
}


static int32_t
ngx_wasm_host_open_argv_env(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    ngx_wasm_call_env_t      *ea;
    ngx_wasm_call_env_ctx_t  *ctx;

    ea = ngx_wasm_host_get_api_context(self, ph);
    if (ea == NULL) {
        return NGX_ERROR;
    }

    ctx = ngx_pcalloc(self->host->pool, sizeof(ngx_wasm_call_env_ctx_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ctx->result = ea->result;
    ctx->state = ea->state;

    ngx_str_array_init_array_encoder(&ctx->sa, ea->args->nelts, ea->args->elts);

    self->data = ctx;

    return NGX_OK;
}


static int32_t
ngx_wasm_host_open_void_env(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    ngx_wasm_call_env_t      *ea;
    ngx_wasm_call_env_ctx_t  *ctx;

    ea = ngx_wasm_host_get_api_context(self, ph);
    if (ea == NULL) {
        return NGX_ERROR;
    }

    ctx = ngx_pcalloc(self->host->pool, sizeof(ngx_wasm_call_env_ctx_t));
    if (ctx == NULL) {
        return NGX_ERROR;
    }

    ctx->result = ea->result;
    ctx->state = ea->state;

    self->data = ctx;

    return NGX_OK;
}


static int32_t
ngx_wasm_host_read_kv_env(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_call_env_ctx_t  *ce = self->data;

    uint8_t  *p;

    p = ngx_encode_array(&ce->kva.arr, &ce->kva.ai, buf, buf + len);
    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


static int32_t
ngx_wasm_host_read_argv_env(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_call_env_ctx_t *ce = self->data;

    uint8_t  *p;

    p = ngx_encode_array(&ce->sa.arr, &ce->sa.ai, buf, buf + len);
    if (p == NULL) {
        return NGX_ERROR;
    }

    return p - buf;
}


static int32_t
ngx_wasm_call_env_set(ngx_wasm_handle_t *self, uint32_t propid, uint8_t *buf,
    uint32_t len)
{
    ngx_wasm_call_env_ctx_t *ce = self->data;

    uint32_t   *in;
    ngx_str_t  *res;

    switch (propid) {

    case NGX_WASI_CALL_PROP_RES:

        res = ce->result;

        res->len = len;

        if (len == 0) {
            res->data= NULL;
            return NGX_OK;
        }

        res->data = ngx_pnalloc(self->host->pool, len);
        if (res->data == NULL) {
            return NGX_ERROR;
        }

        ngx_memcpy(res->data, buf, len);

        return NGX_OK;

    case NGX_WASI_CALL_PROP_STATE:

        if (len != (sizeof(uint32_t))) {
            return NGX_ERROR;
        }

        if (ce->state == NULL) {
            return NGX_ERROR;
        }

        in = (uint32_t *) buf;

        *ce->state = *in;

        return NGX_OK;

    default:
        return NGX_ERROR;
    }

    return NGX_ERROR;
}


static int32_t
ngx_wasm_call_env_get(ngx_wasm_handle_t *self, uint32_t propid, uint8_t *buf,
    uint32_t len)
{
    ngx_wasm_call_env_ctx_t *ce = self->data;

    uint32_t  *out;

    if (propid != NGX_WASI_CALL_PROP_STATE) {
        return NGX_ERROR;
    }

    if (len != (sizeof(uint32_t))) {
        return NGX_ERROR;
    }

    if (ce->state == NULL) {
        return NGX_ERROR;
    }

    out = (uint32_t *) buf;

    *out = *ce->state;

    return NGX_OK;
}
