
/*
 * Copyright (C) 2024 Web Server LLC
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_event_connect.h>

#include <ngx_wasm.h>
#include <ngx_wasm_event.h>
#include <ngx_wasm_connection.h>
#include <wasi/ngx_wasi_connection.h>


#define NGX_WASM_CONNECTION_TAG 0x434F4E4E /* CONN */

typedef struct  {
    ngx_peer_connection_t  peer;
    ngx_str_t              dst;
    ngx_url_t              url;
    ngx_msec_t             connect_timeout;
} ngx_wasm_connection_t;


typedef struct {
    ngx_wasm_connection_t  *connection;
    ngx_wasm_host_t        *host;
} ngx_wasm_connection_cleanup_t;


static int32_t ngx_wasm_host_open_conn(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *parent);
static int32_t ngx_wasm_host_read_conn(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);
static int32_t ngx_wasm_host_write_conn(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);
static int32_t ngx_wasm_host_set_property_conn(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_host_close_conn(ngx_wasm_handle_t *self);
static void ngx_wasm_host_cleanup_connection(void *data);
static void ngx_wasm_connection_close(ngx_wasm_connection_t *wc);


ngx_wasm_handle_ops_t ngx_wasm_connection_ops[] = {
    {   .type = ngx_string("ngx::core::connection"),
        .tag = NGX_WASM_CONNECTION_TAG,
        .open = ngx_wasm_host_open_conn,
        .read = ngx_wasm_host_read_conn,
        .write = ngx_wasm_host_write_conn,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_set_property_conn,
        .close = ngx_wasm_host_close_conn,
    },

    ngx_wasm_host_null_ops
};


static int32_t
ngx_wasm_host_open_conn(ngx_wasm_handle_t *self, ngx_wasm_handle_t *parent)
{
    ngx_pool_cleanup_t             *cln;
    ngx_wasm_connection_t          *wc;
    ngx_wasm_connection_cleanup_t  *wcc;

    wc = ngx_pcalloc(self->host->pool, sizeof(ngx_wasm_connection_t));
    if (wc == NULL) {
        return NGX_ERROR;
    }

    cln = ngx_pool_cleanup_add(self->host->pool,
                               sizeof(ngx_wasm_connection_cleanup_t));
    if (cln == NULL) {
        return NGX_ERROR;
    }

    cln->handler = ngx_wasm_host_cleanup_connection;
    wcc = cln->data;
    wcc->connection = wc;
    wcc->host = self->host;

    self->data = wc;

    return NGX_OK;
}


static int32_t
ngx_wasm_host_read_conn(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_connection_t *wc = self->data;

    ssize_t            n;
    ngx_wasm_event_t  *wasm_ev;
    ngx_connection_t  *c;

    c = wc->peer.connection;

    n = ngx_recv(c, buf, len);

    if (c->read->eof) {

        wasm_ev = ngx_wasm_host_alloc_event(self->host);
        if (wasm_ev == NULL) {
            return NGX_ERROR;
        }

        wasm_ev->fd = self->fd;
        wasm_ev->ptr = self->rev_handler;

        wasm_ev->flags |= NGX_WASM_EV_EOF;

        ngx_queue_insert_tail(&self->host->events, &wasm_ev->queue);
        ngx_post_event(self->host->fd_event, &ngx_posted_events);
    }

    if (n == NGX_AGAIN) {
        if (ngx_handle_read_event(c->read, 0) != NGX_OK) {
            return NGX_ERROR;
        }
    }


    return n;
}


static int32_t
ngx_wasm_host_write_conn(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_connection_t *wc = self->data;

    ssize_t            n;
    ngx_connection_t  *c;

    c = wc->peer.connection;

    n = ngx_send(c, buf, len);

    if (n == NGX_AGAIN) {
        if (ngx_handle_write_event(c->write, 0) != NGX_OK) {
            return NGX_ERROR;
        }
    }

    return n;
}


static void
ngx_wasm_connection_event_handler(ngx_event_t *ev, ngx_uint_t is_write)
{
    ngx_connection_t   *c;
    ngx_wasm_event_t   *wasm_ev;
    ngx_wasm_handle_t  *self;

    c = ev->data;
    self = c->data;

    ngx_log_debug5(NGX_LOG_DEBUG_EVENT, ev->log, 0,
                   "%V: connection %s handler ready:%d err:%d eof:%d",
                   &self->host->name, is_write ? "write" : "read",
                   ev->ready, ev->error, ev->eof);

    wasm_ev = ngx_wasm_host_alloc_event(self->host);
    if (wasm_ev == NULL) {
        return;
    }

    wasm_ev->fd = self->fd;
    wasm_ev->ptr = is_write ? self->wev_handler : self->rev_handler;

    if (ev->timedout) {
        wasm_ev->flags |= NGX_WASM_EV_TIMEDOUT;
    }

    if (ev->error) {
        wasm_ev->flags |= NGX_WASM_EV_ERROR;
    }

    if (ev->eof) {
        wasm_ev->flags |= NGX_WASM_EV_EOF;
    }

    ngx_queue_insert_tail(&self->host->events, &wasm_ev->queue);

    /* will trigger wasm event processing */
    ngx_post_event(self->host->fd_event, &ngx_posted_events);
}


static void
ngx_wasm_connection_read_handler(ngx_event_t *rev)
{
    ngx_wasm_connection_event_handler(rev, 0);
}


static void
ngx_wasm_connection_write_handler(ngx_event_t *wev)
{
    ngx_wasm_connection_event_handler(wev, 1);
}


static int32_t
ngx_wasm_connection_setup(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_connection_t *wc = self->data;

    ngx_addr_t  *addr;

    if (len == 0) {
        return NGX_ERROR;
    }

    wc->dst.len = len;

    wc->dst.data = ngx_pnalloc(self->host->pool, len);
    if (wc->dst.data == NULL) {
        return NGX_ERROR;
    }

    ngx_memcpy(wc->dst.data, buf, len);

    ngx_memzero(&wc->url, sizeof(ngx_url_t));

    wc->url.url = wc->dst;
    wc->url.uri_part = 1;
    wc->url.no_resolve = 1;

    if (ngx_parse_url(self->host->pool, &wc->url) != NGX_OK) {
        if (wc->url.err) {
            ngx_log_error(NGX_LOG_ERR, self->log, 0,
                          "%s in conn \"%V\"", wc->url.err, &wc->url.url);
        }

        return NGX_ERROR;
    }

    addr = &wc->url.addrs[0];

    wc->peer.sockaddr = addr->sockaddr;
    wc->peer.socklen = addr->socklen;
    wc->peer.name = &addr->name;
    wc->peer.get = ngx_event_get_peer;
    wc->peer.log = self->log;
    wc->peer.log_error = NGX_ERROR_ERR;

    return len;
}


static int32_t
ngx_wasm_connection_connect(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_connection_t *wc = self->data;

    ngx_int_t          rc;
    ngx_pool_t        *pool;
    ngx_connection_t  *c;

    pool = ngx_create_pool(2048, self->host->log);
    if (pool == NULL) {
        return NGX_ERROR;
    }

    rc = ngx_event_connect_peer(&wc->peer);

    if (rc == NGX_ERROR || rc == NGX_DECLINED) {
        ngx_destroy_pool(pool);
        return NGX_ERROR;
    }

    c = wc->peer.connection;

    c->data = self;
    c->pool = pool;

    c->read->handler = ngx_wasm_connection_read_handler;
    c->write->handler = ngx_wasm_connection_write_handler;

    if (wc->connect_timeout) {
        ngx_add_timer(c->write, wc->connect_timeout);
    }

    if (rc == NGX_OK) {
        ngx_wasm_connection_write_handler(wc->peer.connection->write);
        return NGX_OK;
    }

    return NGX_AGAIN;
}


static int32_t
ngx_wasm_connection_set_handlers(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    uint32_t  *ch;

    if (len != sizeof(uint32_t) * 2) {
        return NGX_ERROR;
    }

    ch = (uint32_t *) buf;

    self->rev_handler = ch[0];
    self->wev_handler = ch[1];

    return 0;
}

static int32_t
ngx_wasm_connection_set_timeouts(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len)
{
    ngx_wasm_connection_t *wc = self->data;

    ngx_connection_t  *c;

    uint32_t *to;

    if (len != sizeof(uint32_t)) {
        return NGX_ERROR;
    }

    to = (uint32_t *) buf;

    if (wc->peer.connection == NULL) {

        if (propid == NGX_WASI_CONNECTION_OPS_SET_CONNECT_TIMEOUT) {
            wc->connect_timeout = to[0];
            return 0;
        }

        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V not connection yet to set timeouts",
                      &self->host->name);
        return -1;
    }

    c = wc->peer.connection;

    switch (propid) {
    case NGX_WASI_CONNECTION_OPS_SET_CONNECT_TIMEOUT:
    case NGX_WASI_CONNECTION_OPS_SET_READ_TIMEOUT:

        if (to[0] == 0) {
            ngx_del_timer(c->read);

        } else {
            ngx_add_timer(c->read, to[0]);
        }

        return 0;

    case NGX_WASI_CONNECTION_OPS_SET_WRITE_TIMEOUT:
        if (to[0] == 0) {
            ngx_del_timer(c->write);

        } else {
            ngx_add_timer(c->write, to[0]);
        }
        return 0;

    default:
        ngx_log_error(NGX_LOG_ERR, self->log, 0,
                      "%V unknown timeout id", &self->host->name);
        return -1;
    }

    return 0;
}


static int32_t
ngx_wasm_host_set_property_conn(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len)
{
    ngx_log_debug2(NGX_LOG_DEBUG_CORE, self->log, 0,
                   "%V: set property %D", &self->host->name, propid);

    switch (propid) {
    case NGX_WASI_CONNECTION_OPS_SET_ADDR:
        return ngx_wasm_connection_setup(self, buf, len);

    case NGX_WASI_CONNECTION_OPS_SET_HANDLERS:
        return ngx_wasm_connection_set_handlers(self, buf, len);

    case NGX_WASI_CONNECTION_OPS_CONNECT:
        return ngx_wasm_connection_connect(self, buf, len);

    case NGX_WASI_CONNECTION_OPS_SET_CONNECT_TIMEOUT:
    case NGX_WASI_CONNECTION_OPS_SET_READ_TIMEOUT:
    case NGX_WASI_CONNECTION_OPS_SET_WRITE_TIMEOUT:
        return ngx_wasm_connection_set_timeouts(self, propid, buf, len);

    default:
        return NGX_ERROR;
    }
}


static int32_t
ngx_wasm_host_close_conn(ngx_wasm_handle_t *self)
{
    ngx_wasm_connection_close(self->data);

    return NGX_OK;
}


static void
ngx_wasm_host_cleanup_connection(void *data)
{
    ngx_wasm_connection_cleanup_t *wcc = data;

    ngx_connection_t       *c;
    ngx_wasm_connection_t  *wc;

    wc = wcc->connection;

    c = wc->peer.connection;

    if (c == NULL) {
        return;
    }

#if (NGX_DEBUG)
    {
    ngx_wasm_host_t  *wh;

    wh = wcc->host;

    /* TODO: log peer name - may be missing*/
    ngx_log_debug2(NGX_LOG_DEBUG_CORE, wh->log, 0,
                   "%V host: cleanup connection %p", &wh->name, wc);
    }
#endif

    ngx_wasm_connection_close(wc);
}


static void
ngx_wasm_connection_close(ngx_wasm_connection_t *wc)
{
    ngx_pool_t  *pool;

    pool = wc->peer.connection->pool;

    ngx_close_connection(wc->peer.connection);

    if (pool) {
        ngx_destroy_pool(pool);
    }

    wc->peer.connection = NULL;
}

