
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_wasm.h>
#include <ngx_wasm_filter.h>


#define NGX_WASM_HOST_FILTER_OPS_TAG 0x464C5452 /* FLTR */
#define NGX_WASM_FILTER_BUFFER_SIZE  4096


/* filter state */
typedef struct {
    ngx_pool_t        *pool;
    ngx_chain_t       *in;
    ngx_chain_t       *end;
    ngx_chain_t      **last;
    ngx_chain_t      **freep;
    off_t              off;
    ngx_buf_tag_t      tag;
} ngx_wasm_filter_chain_ctx_t;


static int32_t ngx_wasm_filter_open(ngx_wasm_handle_t *self,
    ngx_wasm_handle_t *parent);
static int32_t ngx_wasm_filter_close(ngx_wasm_handle_t *self);

static int32_t ngx_wasm_filter_read(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);
static int32_t ngx_wasm_filter_write(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len);

static size_t ngx_chain_to_buffer(ngx_chain_t *in, off_t off, uint8_t *buf,
    size_t size);


ngx_wasm_handle_ops_t ngx_wasm_filter_ops[] = {
    {   .type = ngx_string("ngx::filter"),
        .tag = NGX_WASM_HOST_FILTER_OPS_TAG,
        .open = ngx_wasm_filter_open,
        .read = ngx_wasm_filter_read,
        .write = ngx_wasm_filter_write,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
        .close = ngx_wasm_filter_close,
    },

    ngx_wasm_host_null_ops
};


static int32_t
ngx_wasm_filter_open(ngx_wasm_handle_t *self, ngx_wasm_handle_t *ph)
{
    ngx_pool_t                   *pool;
    ngx_wasm_filter_env_t        *env;
    ngx_wasm_filter_chain_ctx_t  *cc;

    pool = self->host->pool;

    env = ngx_wasm_host_get_api_context(self, ph);
    if (env == NULL) {
        return NGX_ERROR;
    }

    cc = ngx_pcalloc(pool, sizeof(ngx_wasm_filter_chain_ctx_t));
    if (cc == NULL) {
        return NGX_ERROR;
    }

    cc->pool = pool;
    cc->in = env->in;
    cc->last = env->outp;
    cc->freep = env->freep;
    cc->tag = env->tag;
    /*
     * set by ngx_pcalloc():
     *
     * cc->off = 0;
     * cc->out = NULL;
     * cc->end = NULL;
     */

    self->data = cc;

    return NGX_OK;
}


static int32_t
ngx_wasm_filter_close(ngx_wasm_handle_t *self)
{
    ngx_wasm_filter_chain_ctx_t *cc = self->data;

    ngx_chain_t  *cl;

    cl = cc->end;
    if (cl) {
        cl->buf->last_buf = 1;
        cl->next = NULL;
    }

    return NGX_OK;
}


static int32_t
ngx_wasm_filter_read(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_filter_chain_ctx_t *cc = self->data;

    size_t  n;

    n = ngx_chain_to_buffer(cc->in, cc->off, buf, len);

    cc->off += n;

    return n;
}


static int32_t
ngx_wasm_filter_write(ngx_wasm_handle_t *self, uint8_t *buf, uint32_t len)
{
    ngx_wasm_filter_chain_ctx_t *cc = self->data;

    size_t        bsize;
    uint32_t      left;
    ngx_buf_t    *b;
    ngx_chain_t  *tl, **ll;

    if (len == 0) {
        return 0;
    }

    left = len;
    ll = cc->last;
    tl = NULL;

    /* fill the last buffer */
    if (cc->end && cc->end->buf && cc->end->buf->start) {
        b = cc->end->buf;

        bsize = b->end - b->last;

        if (bsize == 0) {
            goto next_buf;
        }

        if (bsize >= len) {
            ngx_memcpy(b->last, buf, len);
            b->last += len;

            return len;

        } else {
            ngx_memcpy(b->last, buf, bsize);
            b->last += bsize;
            buf += bsize;
            left -= bsize;
        }
    }

next_buf:

    /* add more buffers to chain */
    while (left) {

        tl = ngx_chain_get_free_buf(cc->pool, cc->freep);
        if (tl == NULL) {
            return NGX_ERROR;
        }

        b = tl->buf;
        b->tag = cc->tag;
        b->memory = 1;

        if (b->start == NULL) {

            bsize = ngx_max(len, NGX_WASM_FILTER_BUFFER_SIZE);

            b->start = ngx_pcalloc(cc->pool, bsize);
            if (b->start == NULL) {
                return NGX_ERROR;
            }
            b->end = b->start + bsize;
            b->pos = b->last = b->start;

        } else {
            bsize = b->end - b->start;
            b->pos = b->last = b->start;
        }

        *ll = tl;
        ll = &tl->next;

        if (left <= bsize) {

            ngx_memcpy(b->pos, buf, left);
            b->last += left;

            break;
        }

        ngx_memcpy(b->pos, buf, bsize);
        b->last += bsize;

        buf += bsize;
        left -= bsize;
    }

    cc->last = ll;
    cc->end = tl;

    return len;
}


static size_t
ngx_chain_to_buffer(ngx_chain_t *in, off_t off, uint8_t *buf, size_t size)
{
    size_t        n, left, len;
    uint8_t      *pos, *rpos,skip;
    ngx_buf_t    *b;
    ngx_chain_t  *cl;

    n = 0;
    pos = buf;

    if (off) {
        skip = 1;
        left = off;

    } else {
        skip = 0;
        left = size;
    }

    for (cl = in; cl; cl = cl->next) {

        b = cl->buf;

        if (ngx_buf_special(b)) {
            continue;
        }

        len = ngx_buf_size(b);

        if (len < left) {

            left -= len;

            if (!skip) {
                pos = ngx_cpymem(pos, b->pos, len);
                n += len;
                b->pos += len;
            }

            continue;
        }

        /* len >= left */

        if (skip) {
            skip = 0;
            len -= left;
            rpos = b->pos +left;

            left = size;

            if (len < left) {
                left -= len;

                ngx_memcpy(pos, rpos, len);
                n += len;
                b->pos += len;

                continue;

            } else {

                ngx_memcpy(pos, rpos, left);
                n += left;
                b->pos += left;

                break;
            }

        } else {

            ngx_memcpy(pos, b->pos, left);
            n += left;
            b->pos += left;

            break;
        }
    }

    return n;
}
