
/*
 * Copyright (C) 2024 Web Server LLC
 */


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_string.h>
#include <ngx_wasm.h>


/* rbtree of open handles with assigned descriptors */
struct ngx_wasm_handles_s {
    ngx_rbtree_t              rbtree;
    ngx_rbtree_node_t         sentinel;

    ngx_pool_t               *pool;

    ngx_queue_t               free;            /* closed handles */
    ngx_uint_t                num;             /* descriptor counter */
};


static void
ngx_wasm_handle_rbtree_insert(ngx_rbtree_node_t *temp, ngx_rbtree_node_t *node,
    ngx_rbtree_node_t *sentinel)
{
    ngx_rbtree_node_t  **p;
    ngx_wasm_handle_t   *n, *nt;

    for ( ;; ) {

        if (node->key < temp->key) {

            p = &temp->left;

        } else if (node->key > temp->key) {

            p = &temp->right;

        } else { /* node->key == temp->key */

            n = (ngx_wasm_handle_t *) node;
            nt = (ngx_wasm_handle_t *) temp;

            p = (n->fd < nt->fd) ? &temp->left : &temp->right;
        }

        if (*p == sentinel) {
            break;
        }

        temp = *p;
    }

    *p = node;
    node->parent = temp;
    node->left = sentinel;
    node->right = sentinel;
    ngx_rbt_red(node);
}


ngx_wasm_handles_t *
ngx_wasm_handles_create(ngx_pool_t *pool)
{
    ngx_wasm_handles_t  *ht;

    ht = ngx_pcalloc(pool, sizeof(ngx_wasm_handles_t));
    if (ht == NULL) {
        return NULL;
    }

    ngx_queue_init(&ht->free);

    /*
     * reserve 0, 1, 2 for stdin, stdout and stderr (unused)
     * reserve 3 for default logger (opened while host created)
     *
     */
    ht->num = 3;

    ngx_rbtree_init(&ht->rbtree, &ht->sentinel, ngx_wasm_handle_rbtree_insert);

    ht->pool = pool;

    return ht;
}


ngx_wasm_handle_t *
ngx_wasm_create_handle(ngx_wasm_handles_t *handles)
{
    ngx_queue_t        *q;
    ngx_wasm_handle_t  *handle;
    ngx_rbtree_node_t  *node;

    if (ngx_queue_empty(&handles->free)) {
        handle = ngx_pcalloc(handles->pool, sizeof(ngx_wasm_handle_t));
        if (handle == NULL) {
            return NULL;
        }

        node = &handle->node;

        node->key = handles->num++;
        handle->fd = node->key;

        ngx_rbtree_insert(&handles->rbtree, node);

    } else {
        q = ngx_queue_head(&handles->free);
        handle = ngx_queue_data(q, ngx_wasm_handle_t, queue);
        ngx_queue_remove(q);

        /* key is reset when node was removed from tree */
        handle->node.key = handle->fd;
        ngx_rbtree_insert(&handles->rbtree, &handle->node);
    }

    return handle;
}


ngx_wasm_handle_t *
ngx_wasm_find_handle(ngx_wasm_handles_t *handles, int32_t fd)
{
    ngx_wasm_handle_t  *n;
    ngx_rbtree_node_t  *node, *sentinel;

    node = handles->rbtree.root;
    sentinel = handles->rbtree.sentinel;

    if (fd < 0) {
        return NULL;
    }

    while (node != sentinel) {

        n = (ngx_wasm_handle_t *) node;

        if ((uint32_t) fd < n->fd) {
            node = node->left;
            continue;
        }

        if ((uint32_t) fd > n->fd) {
            node = node->right;
            continue;
        }

        /* fd == n->fd */
        return n;
    }

    return NULL;
}


void
ngx_wasm_delete_handle(ngx_wasm_handles_t *handles, ngx_wasm_handle_t *handle)
{
    ngx_rbtree_node_t  *node;

    node = &handle->node;

    ngx_rbtree_delete(&handles->rbtree, node);

    ngx_queue_insert_head(&handles->free, &handle->queue);
}

