
/*
 * Copyright (C) 2024 Web Server LLC
 */


#include <ngx_config.h>
#include <ngx_core.h>

#include <ngx_wasm_log.h>
#include <ngx_wasm_event.h>


#define ngx_wasm_host_find_opsz(wh, api)                                      \
    ngx_wasm_host_find_ops(wh, (u_char *) api, ngx_strlen(api));


static void ngx_wasm_host_cleanup(void *data);

static int32_t ngx_wasm_host_open(ngx_wasm_host_t *host, uint8_t *namebuf,
    uint32_t len, int32_t ctx);
static int32_t ngx_wasm_host_close(ngx_wasm_host_t *host, int32_t fd);
static int32_t ngx_wasm_host_read(ngx_wasm_host_t *host, int32_t fd,
    uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_host_write(ngx_wasm_host_t *host, int32_t fd,
    uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_host_get(ngx_wasm_host_t *host, int32_t fd,
    uint32_t propid, uint8_t *buf, uint32_t len);
static int32_t ngx_wasm_host_set(ngx_wasm_host_t *host, int32_t fd,
    uint32_t propid, uint8_t *buf, uint32_t len);

static ngx_inline ngx_wasm_handle_ops_t *ngx_wasm_host_find_ops(
    ngx_wasm_host_t *wh, u_char *name, size_t len);
static ngx_wasm_handle_t *ngx_wasm_host_open_new_handle(ngx_wasm_host_t *host,
    ngx_wasm_handle_ops_t *ops, ngx_wasm_handle_t *ph);
static ngx_inline ngx_wasm_handle_t *ngx_wasm_host_create_api_context_handle(
    ngx_wasm_host_t *wh, ngx_wasm_handle_ops_t *ops, void *data);


static ngx_wasm_host_ops_t ngx_wasm_dispatch_ops = {
    ngx_wasm_host_open,
    ngx_wasm_host_close,
    ngx_wasm_host_read,
    ngx_wasm_host_write,
    ngx_wasm_host_get,
    ngx_wasm_host_set,
};


static ngx_wasm_handle_ops_t ngx_wasm_host_stub_ops[] = {
    {   .type = ngx_string("host_object"),
        .tag = (ngx_uint_t) -1,
        .open = ngx_wasm_host_open_stub,
        .read = ngx_wasm_host_io_stub,
        .write = ngx_wasm_host_io_stub,
        .get = ngx_wasm_host_property_stub,
        .set = ngx_wasm_host_property_stub,
        .close = ngx_wasm_host_close_stub,
    },

    ngx_wasm_host_null_ops
};


ngx_wasm_host_t *
ngx_wasm_host_create(ngx_wasm_instance_t *wi, ngx_pool_t *pool, size_t size,
    ngx_log_t *log, const char *name, ngx_array_t *apis, ngx_array_t *acl)
{
    int32_t                    lfd;
    ngx_str_t                 *items;
    ngx_uint_t                 i, j;
    ngx_pool_t                *tmp_pool;
    ngx_hash_init_t            hash;
    ngx_wasm_host_t           *wh;
    ngx_pool_cleanup_t        *cln;
    ngx_wasm_handle_ops_t     *ops, **api;
    ngx_hash_keys_arrays_t    *hk;

    wh = ngx_pcalloc(pool, size);
    if (wh == NULL) {
        return NULL;
    }

    cln = ngx_pool_cleanup_add(pool, 0);
    if (cln == NULL) {
        return NULL;
    }

    wh->wi = wi;
    wh->pool = pool;
    wh->log = log;
    wh->ops = &ngx_wasm_dispatch_ops;

    wh->name.len = ngx_strlen(name);
    wh->name.data = ngx_pnalloc(pool, wh->name.len);
    if (wh->name.data == NULL) {
        return NULL;
    }
    ngx_memcpy(wh->name.data, name, wh->name.len);

    wh->handles = ngx_wasm_handles_create(pool);
    if (wh->handles == NULL) {
        return NULL;
    }

    ngx_queue_init(&wh->events);
    ngx_queue_init(&wh->free_events);

    tmp_pool = ngx_create_pool(NGX_DEFAULT_POOL_SIZE, pool->log);
    if (tmp_pool == NULL) {
        return NULL;
    }

    hash.hash = &wh->handle_ops;
    hash.key = ngx_hash_key;
    hash.max_size = 512;
    hash.bucket_size = ngx_align(64, ngx_cacheline_size);
    hash.name = "wasm ops hash";
    hash.pool = pool;
    hash.temp_pool = tmp_pool;

    hk = &wh->handle_ops_keys;

    hk->pool = pool;
    hk->temp_pool = tmp_pool;

    ngx_hash_keys_array_init(hk, NGX_HASH_SMALL);

    api = apis->elts;

    for (i = 0; i < apis->nelts; i++) {

        ops = api[i];

        if (acl == NULL) {
            ngx_hash_add_key(hk, &ops->type, ops, NGX_HASH_READONLY_KEY);
            continue;
        }

        items = acl->elts;

        for (j = 0; j < acl->nelts; j++) {
            if (items[j].len == ops->type.len
                && ngx_strncmp(items[j].data, ops->type.data, ops->type.len)
                   == 0)
            {
                ngx_log_debug2(NGX_LOG_DEBUG_CORE, log, 0,
                               "%V host: enabled API \"%V\"",
                               &wh->name, &items[j]);

                ngx_hash_add_key(hk, &ops->type, ops, NGX_HASH_READONLY_KEY);
                break;
            }
        }
    }

    ngx_hash_init(&hash, hk->keys.elts, hk->keys.nelts);

    ngx_destroy_pool(tmp_pool);

    cln->data = wh;
    cln->handler = ngx_wasm_host_cleanup;

    lfd = ngx_wasm_host_create_object(wh, "ngx::core::log", wh->log);
    if (lfd != 3) {
        ngx_log_error(NGX_LOG_ERR, log, 0,
                      "failed to create default log handle");
        return NULL;
    }

    return wh;
}


static void
ngx_wasm_host_cleanup(void *data)
{
    ngx_wasm_host_t  *wh;

    wh = data;

    ngx_log_debug1(NGX_LOG_DEBUG_CORE, wh->log, 0,
                   "%V host: cleanup", &wh->name);

    if (wh->fd_event && wh->fd_event->posted) {
        ngx_delete_posted_event(wh->fd_event);
    }
}


ngx_wasm_event_t *
ngx_wasm_host_alloc_event(ngx_wasm_host_t *host)
{
    ngx_queue_t       *q;
    ngx_wasm_event_t  *ev;

    if (!ngx_queue_empty(&host->free_events)) {
        q = ngx_queue_head(&host->free_events);
        ngx_queue_remove(q);

        ev = ngx_queue_data(q, ngx_wasm_event_t, queue);
        ngx_memzero(ev, sizeof(ngx_wasm_event_t));

    } else {
        ev = ngx_pcalloc(host->pool, sizeof(ngx_wasm_event_t));
        if (ev == NULL) {
            return NULL;
        }
    }

    return ev;
}


int32_t
ngx_wasm_host_create_api_context(ngx_wasm_host_t *wh, const char *api,
    void *data)
{
    ngx_wasm_handle_t      *handle;
    ngx_wasm_handle_ops_t  *ops;

    ops = ngx_wasm_host_find_opsz(wh, api);
    if (ops == NULL) {
        return NGX_ERROR;
    }

    handle = ngx_wasm_host_create_api_context_handle(wh, ops, data);
    if (handle == NULL) {
        return NGX_ERROR;
    }

    return handle->fd;
}


void *
ngx_wasm_host_get_api_context(ngx_wasm_handle_t *handle, ngx_wasm_handle_t *ph)
{
    return ngx_wasm_host_get_object_context(handle, ph,
                                            handle->tag | NGX_WASM_API_CONTEXT);
}


void *
ngx_wasm_host_get_object_context(ngx_wasm_handle_t *handle,
    ngx_wasm_handle_t *ph, ngx_uint_t tag)
{
   if (ph == NULL) {
        ngx_log_error(NGX_LOG_ERR, handle->log, 0,
                      "%V: context handle is missing", &handle->host->name);
        return NULL;
    }

    if (ph->tag != tag) {
        ngx_log_error(NGX_LOG_ERR, handle->log, 0,
                      "%V invalid context handle type (expect 0x%xD got 0x%xi)",
                      &handle->host->name, tag, ph->tag);
        return NULL;
    }

    return ph->data;
}


int32_t
ngx_wasm_host_open_object(ngx_wasm_host_t *wh, const char *api, void *data)
{
    ngx_wasm_handle_t       ph, *handle;
    ngx_wasm_handle_ops_t  *ops;

    ops = ngx_wasm_host_find_opsz(wh, api);
    if (ops == NULL) {
        ngx_log_error(NGX_LOG_ERR, wh->log, 0,
                      "%V: faled to lookup ops for api \"%s\"",
                      &wh->name, api);
        return NGX_ERROR;
    }

    ngx_memzero(&ph, sizeof(ngx_wasm_handle_t));

    ph.tag = ops->tag | NGX_WASM_API_CONTEXT;
    ph.ops = ngx_wasm_host_stub_ops;
    ph.host = wh;
    ph.log = wh->log;
    ph.data = data;
    ph.fd = -1;

    handle = ngx_wasm_host_open_new_handle(wh, ops, &ph);
    if (handle == NULL) {
        return NGX_ERROR;
    }

    return handle->fd;
}


ngx_int_t
ngx_wasm_host_open_as(ngx_wasm_host_t *wh, ngx_wasm_handle_t *handle,
    const char *api, void *data)
{
    ngx_wasm_handle_t       ph;
    ngx_wasm_handle_ops_t  *ops;

    ops = ngx_wasm_host_find_opsz(wh, api);
    if (ops == NULL) {
        ngx_log_error(NGX_LOG_ERR, wh->log, 0,
                      "%V: faled to lookup ops for api \"%s\"",
                      &wh->name, api);
        return NGX_ERROR;
    }

    ngx_memzero(&ph, sizeof(ngx_wasm_handle_t));

    ph.tag = ops->tag | NGX_WASM_API_CONTEXT;
    ph.ops = ngx_wasm_host_stub_ops;
    ph.host = wh;
    ph.log = wh->log;
    ph.data = data;
    ph.fd = -1;

    handle->tag = ops->tag;
    handle->ops = ops;
    handle->host = wh;
    handle->log = wh->log;

    if (ops->open(handle, &ph) != NGX_OK) {
        return NGX_ERROR;
    }

    return NGX_OK;
}


ngx_int_t
ngx_wasm_host_delete_object(ngx_wasm_host_t *host, int32_t fd)
{
    ngx_wasm_handle_t  *handle;

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to delete object",
                      &host->name, fd);
        return NGX_ERROR;
    }

    ngx_wasm_delete_handle(host->handles, handle);

    return NGX_OK;
}


static int32_t
ngx_wasm_host_open(ngx_wasm_host_t *host, uint8_t *namebuf, uint32_t len,
    int32_t ctx)
{
    ngx_str_t               name;
    ngx_wasm_handle_t      *handle, *other;
    ngx_wasm_handle_ops_t  *ops;

    name.data = namebuf;
    name.len = len;

    ngx_log_debug3(NGX_LOG_DEBUG_CORE, host->log, 0,
                   "%V host: open(\"%V\", fd:%D)", &host->name, &name, ctx);

    if (ctx != -1) {
        other = ngx_wasm_find_handle(host->handles, ctx);

        if (other == NULL) {
            ngx_log_error(NGX_LOG_ERR, host->log, 0,
                          "%V: incorrect fd %D passed to open(\"%V\")",
                          &host->name, ctx, &name);
            return NGX_ERROR;
        }

    } else {
        other = NULL;
    }

    ops = ngx_wasm_host_find_ops(host, name.data, name.len);
    if (ops == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: open unknown type \"%V\"",
                      &host->name, &name);
        return NGX_ERROR;
    }

    handle = ngx_wasm_host_open_new_handle(host, ops, other);
    if (handle == NULL) {
        return NGX_ERROR;
    }

    ngx_log_debug4(NGX_LOG_DEBUG_CORE, host->log, 0,
                  "%V host: open(\"%V\",fd:%D) = #%D",
                  &host->name, &name, ctx, handle->fd);

    return handle->fd;
}


int32_t
ngx_wasm_host_create_object(ngx_wasm_host_t *host, const char *api,
    void *data)
{
    ngx_str_t               name;
    ngx_wasm_handle_t      *handle;
    ngx_wasm_handle_ops_t  *ops;

    name.len = ngx_strlen(api);
    name.data = (u_char *) api;

    ops = ngx_wasm_host_find_ops(host, name.data, name.len);
    if (ops == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: open unknown type \"%V\"",
                      &host->name, &name);
        return NGX_ERROR;
    }

    handle = ngx_wasm_create_handle(host->handles);
    if (handle == NULL) {
        return NGX_ERROR;
    }

    handle->tag = ops->tag;
    handle->ops = ops;
    handle->host = host;
    handle->log = host->log;

    /* instead of real open we just set data */
    handle->data = data;

    ngx_log_debug3(NGX_LOG_DEBUG_CORE, host->log, 0,
                  "%V host: create object \"%V\" fd #%D",
                  &host->name, &name, handle->fd);

    return handle->fd;
}


static int32_t
ngx_wasm_host_close(ngx_wasm_host_t *host, int32_t fd)
{
    int32_t             rc;
    ngx_wasm_handle_t  *handle;

    ngx_log_debug2(NGX_LOG_DEBUG_CORE, host->log, 0,
                   "%V host: close(fd:%D)", &host->name, fd);

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to close",
                      &host->name, fd);
        return NGX_ERROR;
    }

    rc = handle->ops->close(handle);

    if (rc == NGX_OK) {
        ngx_wasm_delete_handle(host->handles, handle);
    }

    return rc;
}


static int32_t
ngx_wasm_host_read(ngx_wasm_host_t *host, int32_t fd, uint8_t *buf,
    uint32_t len)
{
    ngx_wasm_handle_t  *handle;

    ngx_log_debug3(NGX_LOG_DEBUG_CORE, host->log, 0,
                   "%V host: read(fd:%D, len:%uD)", &host->name, fd, len);

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to read",
                      &host->name, fd);
        return NGX_ERROR;
    }

    return handle->ops->read(handle, buf, len);
}


static int32_t
ngx_wasm_host_write(ngx_wasm_host_t *host, int32_t fd, uint8_t *buf,
    uint32_t len)
{
    ngx_wasm_handle_t  *handle;

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to write",
                      &host->name, fd);

        return NGX_ERROR;
    }

#if (NGX_DEBUG)
    /* avoid debug for logging itself */
    if (handle->ops->tag != NGX_WASM_HOST_LOG_TAG) {
        ngx_log_debug3(NGX_LOG_DEBUG_CORE, host->log, 0,
                       "%V host: write(fd:%D, len:%D)",
                       &host->name, fd, len);
    }
#endif

    return handle->ops->write(handle, buf, len);
}


static int32_t
ngx_wasm_host_get(ngx_wasm_host_t *host, int32_t fd, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    ngx_wasm_handle_t  *handle;

    ngx_log_debug4(NGX_LOG_DEBUG_CORE, host->log, 0,
                   "%V host: get(fd:%D, id:%D, len:%D)",
                   &host->name, fd, propid, len);

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to get",
                      &host->name, fd);

        return NGX_ERROR;
    }

    return handle->ops->get(handle, propid, buf, len);
}


static int32_t
ngx_wasm_host_set(ngx_wasm_host_t *host, int32_t fd, uint32_t propid,
    uint8_t *buf, uint32_t len)
{
    ngx_wasm_handle_t  *handle;

    ngx_log_debug4(NGX_LOG_DEBUG_CORE, host->log, 0,
                   "%V host: set(fd:%D, id:%D, len:%D)",
                   &host->name, fd, propid, len);

    handle = ngx_wasm_find_handle(host->handles, fd);
    if (handle == NULL) {
        ngx_log_error(NGX_LOG_ERR, host->log, 0,
                      "%V: invalid handle %D passed to set",
                      &host->name, fd);
        return NGX_ERROR;
    }

    return handle->ops->set(handle, propid, buf, len);
}


int32_t
ngx_wasm_host_open_stub(ngx_wasm_handle_t *self, ngx_wasm_handle_t *parent)
{
    return NGX_ERROR;
}


int32_t
ngx_wasm_host_close_stub(ngx_wasm_handle_t *self)
{
    return NGX_OK;
}


int32_t
ngx_wasm_host_io_stub(ngx_wasm_handle_t *self, uint8_t *buf,
    uint32_t len)
{
    return NGX_ERROR;
}


int32_t
ngx_wasm_host_property_stub(ngx_wasm_handle_t *self,
    uint32_t propid, uint8_t *buf, uint32_t len)
{
    return NGX_ERROR;
}


static ngx_inline ngx_wasm_handle_ops_t *
ngx_wasm_host_find_ops(ngx_wasm_host_t *wh, u_char *name, size_t len)
{
    ngx_uint_t  key;

    key = ngx_hash_key(name, len);

    return ngx_hash_find(&wh->handle_ops, key, name, len);
}


static ngx_wasm_handle_t *
ngx_wasm_host_open_new_handle(ngx_wasm_host_t *host,
    ngx_wasm_handle_ops_t *ops, ngx_wasm_handle_t *ph)
{
    ngx_wasm_handle_t  *handle;

    handle = ngx_wasm_create_handle(host->handles);
    if (handle == NULL) {
        return NULL;
    }

    handle->tag = ops->tag;
    handle->ops = ops;
    handle->host = host;
    handle->log = host->log;

    if (ops->open(handle, ph) != NGX_OK) {
        ngx_wasm_delete_handle(host->handles, handle);
        return NULL;
    }

    return handle;
}


static ngx_inline ngx_wasm_handle_t *
ngx_wasm_host_create_api_context_handle(ngx_wasm_host_t *wh,
    ngx_wasm_handle_ops_t *ops, void *data)
{
    ngx_wasm_handle_t  *handle;

    handle = ngx_wasm_create_handle(wh->handles);
    if (handle == NULL) {
        return NULL;
    }

    /* mark the object as special context for this API */
    handle->tag = ops->tag | NGX_WASM_API_CONTEXT;
    handle->ops = ngx_wasm_host_stub_ops;
    handle->host = wh;
    handle->log = wh->log;
    handle->data = data;

    return handle;
}

