
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_wasm.h>


/*
 * Implementation of generic WASM module type.
 * The ngx_event.c was used as a model.  Refer to it if in doubt.
 */


static char *ngx_wasm_plugins_block(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);
static ngx_int_t ngx_wasm_verify_modules(ngx_cycle_t *cycle);

static char *ngx_wasm_core_use(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
static char *ngx_wasm_core_load_module(ngx_conf_t *cf, ngx_command_t *cmd,
    void *conf);

static void *ngx_wasm_core_create_conf(ngx_cycle_t *cycle);
static char *ngx_wasm_core_init_conf(ngx_cycle_t *cycle, void *conf);

static ngx_int_t ngx_wasm_core_module_init(ngx_cycle_t *cycle);
static ngx_int_t ngx_wasm_core_process_init(ngx_cycle_t *cycle);
static void ngx_wasm_core_process_exit(ngx_cycle_t *cycle);


static ngx_command_t  ngx_wasm_commands[] = {

    /* NB: The 'wasm' name is already taken by kong module */
    { ngx_string("wasm_modules"),
      NGX_MAIN_CONF|NGX_CONF_BLOCK|NGX_CONF_NOARGS,
      ngx_wasm_plugins_block,
      0,
      0,
      NULL },


    ngx_null_command
};


static ngx_command_t  ngx_wasm_core_commands[] = {

    { ngx_string("use"),
      NGX_WASM_CONF|NGX_CONF_TAKE1,
      ngx_wasm_core_use,
      0,
      0,
      NULL },

    { ngx_string("load"),
      NGX_WASM_CONF|NGX_CONF_TAKE1|NGX_CONF_1MORE,
      ngx_wasm_core_load_module,
      0,
      0,
      NULL },

      ngx_null_command
};


static ngx_core_module_t  ngx_wasm_module_ctx = {
    ngx_string("WASM"),
    NULL,
    NULL
};

static ngx_str_t ngx_wasm_core_module_name = ngx_string("wasm_core");

static ngx_wasm_core_module_t  ngx_wasm_core_module_ctx = {
    &ngx_wasm_core_module_name,
    ngx_wasm_core_create_conf,
    ngx_wasm_core_init_conf,

    { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 0 }
};


ngx_module_t  ngx_wasm_module = {
    NGX_MODULE_V1,
    &ngx_wasm_module_ctx,             /* module context */
    ngx_wasm_commands,                /* module directives */
    NGX_CORE_MODULE,                  /* module type */
    NULL,                             /* init master */
    NULL,                             /* init module */
    NULL,                             /* init process */
    NULL,                             /* init thread */
    NULL,                             /* exit thread */
    NULL,                             /* exit process */
    NULL,                             /* exit master */
    NGX_MODULE_V1_PADDING
};


ngx_module_t  ngx_wasm_core_module = {
    NGX_MODULE_V1,
    &ngx_wasm_core_module_ctx,        /* module context */
    ngx_wasm_core_commands,           /* module directives */
    NGX_WASM_MODULE,                  /* module type */
    NULL,                             /* init master */
    ngx_wasm_core_module_init,        /* init module */
    ngx_wasm_core_process_init,       /* init process */
    NULL,                             /* init thread */
    NULL,                             /* exit thread */
    ngx_wasm_core_process_exit,       /* exit process */
    NULL,                             /* exit master */
    NGX_MODULE_V1_PADDING
};


ngx_wasm_actions_t    ngx_wasm_actions;
static ngx_uint_t     ngx_wasm_max_module;


static char *
ngx_wasm_plugins_block(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    char                      *rv;
    void                    ***ctx;
    ngx_uint_t                 i;
    ngx_conf_t                 pcf;
    ngx_wasm_core_module_t    *m;

    if (*(void **) conf) {
        return "is duplicate";
    }

    /* count the number of the wasm modules and set up their indices */

    ngx_wasm_max_module = ngx_count_modules(cf->cycle, NGX_WASM_MODULE);

    ctx = ngx_pcalloc(cf->pool, sizeof(void *));
    if (ctx == NULL) {
        return NGX_CONF_ERROR;
    }

    *ctx = ngx_pcalloc(cf->pool, ngx_wasm_max_module * sizeof(void *));
    if (*ctx == NULL) {
        return NGX_CONF_ERROR;
    }

    *(void **) conf = ctx;

    for (i = 0; cf->cycle->modules[i]; i++) {
        if (cf->cycle->modules[i]->type != NGX_WASM_MODULE) {
            continue;
        }

        m = cf->cycle->modules[i]->ctx;

        if (m->create_conf) {
            (*ctx)[cf->cycle->modules[i]->ctx_index] =
                                                     m->create_conf(cf->cycle);
            if ((*ctx)[cf->cycle->modules[i]->ctx_index] == NULL) {
                return NGX_CONF_ERROR;
            }
        }
    }

    pcf = *cf;
    cf->ctx = ctx;
    cf->module_type = NGX_WASM_MODULE;
    cf->cmd_type = NGX_WASM_CONF;

    rv = ngx_conf_parse(cf, NULL);

    *cf = pcf;

    if (rv != NGX_CONF_OK) {
        return rv;
    }

    for (i = 0; cf->cycle->modules[i]; i++) {
        if (cf->cycle->modules[i]->type != NGX_WASM_MODULE) {
            continue;
        }

        m = cf->cycle->modules[i]->ctx;

        if (m->init_conf) {
            rv = m->init_conf(cf->cycle,
                              (*ctx)[cf->cycle->modules[i]->ctx_index]);
            if (rv != NGX_CONF_OK) {
                return rv;
            }
        }
    }

    if (ngx_wasm_verify_modules(cf->cycle) != NGX_OK) {
        return NGX_CONF_ERROR;
    }

    return NGX_CONF_OK;
}


/*
 * creates temporary runtime at configuration stage in order
 * to verify modules to be loaded: detect files in bad format,
 * missing functions, etc).
 */
static ngx_int_t
ngx_wasm_verify_modules(ngx_cycle_t *cycle)
{
    ngx_uint_t            i;
    ngx_wasm_conf_t      *wcf;
    ngx_wasm_module_t    *wm;
    ngx_wasm_instance_t  *wi;

    /* some runtimes cannot be stopped, thus avoid them */
    if (!ngx_wasm_actions.validate_on_start) {
        return NGX_OK;
    }

    wcf = ngx_wasm_get_conf(cycle->conf_ctx, ngx_wasm_core_module);

    /* nothing to check */
    if (wcf->modules.nelts == 0) {
        return NGX_OK;
    }

    if (ngx_wasm_create_runtime(cycle) != NGX_OK) {
        return NGX_ERROR;
    }

    wm = wcf->modules.elts;

    for (i = 0; i < wcf->modules.nelts; i++) {

        if (ngx_wasm_create_module(cycle, &wm[i]) != NGX_OK) {
            ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                          "failed to verify module \"%V\"",
                          &wm[i].file);

            return NGX_ERROR;
        }

        /*
         * NB: we are probably running as root here, and module
         * instantiation with WASI enabled may call __initialize__()
         */
        wi = ngx_wasm_create_instance(&wm[i], cycle->pool, cycle->log);
        if (wi == NULL) {
            ngx_log_error(NGX_LOG_EMERG, cycle->log, 0,
                          "failed to verify module \"%V\"",
                          &wm[i].file);

            return NGX_ERROR;
        }

        ngx_log_debug1(NGX_LOG_DEBUG_CORE, cycle->log, 0,
                       "module \"%V\" validated", &wm[i].file);

        ngx_wasm_delete_instance(wi);
        ngx_wasm_delete_module(&wm[i]);
    }

    ngx_wasm_delete_runtime(cycle);

    return NGX_OK;
}


ngx_wasm_module_t *
ngx_wasm_get_module(ngx_conf_t *cf, ngx_str_t *ns)
{
    ngx_uint_t          i;
    ngx_wasm_conf_t    *wcf;
    ngx_wasm_module_t  *wm;

    wcf = (ngx_wasm_conf_t *) ngx_wasm_get_conf(cf->cycle->conf_ctx,
                                                ngx_wasm_core_module);
    if (wcf == NULL) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "wasm_modules block is missing");
        return NULL;
    }

    wm = wcf->modules.elts;

    for (i = 0; i < wcf->modules.nelts; i++) {
        if (wm[i].ns.len != ns->len) {
            continue;
        }

        if (ns->len == 0) {
            return &wm[i];
        }

        if (ngx_strncmp(wm[i].ns.data, ns->data, ns->len) != 0) {
            continue;
        }

        return &wm[i];
    }

    return NULL;
}


static char *
ngx_wasm_core_use(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_wasm_conf_t  *wcf = conf;

    ngx_int_t                 m;
    ngx_str_t                *value;
    ngx_wasm_conf_t          *old_wcf;
    ngx_wasm_core_module_t   *module;

    if (wcf->use != NGX_CONF_UNSET_UINT) {
        return "is duplicate";
    }

    value = cf->args->elts;

    if (cf->cycle->old_cycle->conf_ctx) {
        old_wcf = ngx_wasm_get_conf(cf->cycle->old_cycle->conf_ctx,
                                     ngx_wasm_core_module);
    } else {
        old_wcf = NULL;
    }


    for (m = 0; cf->cycle->modules[m]; m++) {
        if (cf->cycle->modules[m]->type != NGX_WASM_MODULE) {
            continue;
        }

        module = cf->cycle->modules[m]->ctx;
        if (module->name->len == value[1].len) {
            if (ngx_strcmp(module->name->data, value[1].data) == 0) {
                wcf->use = cf->cycle->modules[m]->ctx_index;
                wcf->name = module->name->data;

                if (ngx_process == NGX_PROCESS_SINGLE
                    && old_wcf
                    && old_wcf->use != wcf->use)
                {
                    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                               "when the server runs without a master process "
                               "the \"%V\" wasm type must be the same as "
                               "in previous configuration - \"%s\" "
                               "and it cannot be changed on the fly, "
                               "to change it you need to stop server "
                               "and start it again",
                               &value[1], old_wcf->name);

                    return NGX_CONF_ERROR;
                }

                return NGX_CONF_OK;
            }
        }
    }

    ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                       "invalid wasm type \"%V\"", &value[1]);

    return NGX_CONF_ERROR;
}


static char *
ngx_wasm_core_load_module(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_wasm_conf_t  *wcf = conf;

    u_char              *buf, *p;
    size_t               size;
    ssize_t              n;
    ngx_str_t           *value, *acl;
    ngx_uint_t           i;
    ngx_file_t           file;
    ngx_file_info_t      fi;
    ngx_wasm_module_t    *wm;
    ngx_wasm_preopen_t   *po;

#if defined (NGX_HTTP_WASM)
    ngx_uint_t           *k;
#endif

    buf = NULL;
#if (NGX_SUPPRESS_WARN)
    size = 0;
#endif

    value = cf->args->elts;

    wm = ngx_array_push(&wcf->modules);
    if (wm == NULL) {
        return NGX_CONF_ERROR;
    }

    ngx_memzero(wm, sizeof(ngx_wasm_module_t));

    wm->index = wm - (ngx_wasm_module_t *) (wcf->modules.elts);

    wm->file = value[1];

    if (ngx_conf_full_name(cf->cycle, &value[1], 1) != NGX_OK) {
        return NGX_CONF_ERROR;
    }

    for (i = 2; i < cf->args->nelts; i++) {

        if (ngx_strncmp(value[i].data, "id=", 3) == 0) {
            value[i].data += 3;
            value[i].len -= 3;

            if (value[i].len == 0) {
                return "empty module id";
            }

            wm->ns = value[i];
            continue;
        }

        if (ngx_strncmp(value[i].data, "type=", 5) == 0) {
            value[i].data += 5;
            value[i].len -= 5;

            if (value[i].len == 0) {
                return "empty module type";
            }

            if (value[i].len == 7
                && ngx_strncmp(value[i].data, "reactor", 7) == 0)
            {
                wm->is_reactor = 1;
                continue;
            }

            if (value[i].len == 7
                && ngx_strncmp(value[i].data, "command", 7) == 0)
            {
                wm->is_reactor = 0;
                continue;
            }

            return "unknown module type";
        }

        if (ngx_strncmp(value[i].data, "api=", 4) == 0) {
            value[i].data += 4;
            value[i].len -= 4;

            if (wm->acl == NULL) {
                wm->acl = ngx_array_create(cf->pool, 4, sizeof(ngx_str_t));
                if (wm->acl == NULL) {
                    return NGX_CONF_ERROR;
                }
            }

            acl = ngx_array_push(wm->acl);
            if (acl == NULL) {
                return NGX_CONF_ERROR;
            }

            *acl = value[i];

            continue;
        }

        if (ngx_strncmp(value[i].data, "fs=", 3) == 0) {
            value[i].data += 3;
            value[i].len -= 3;

            if (wm->preopen == NULL) {
                wm->preopen = ngx_array_create(cf->pool, 4,
                                           sizeof(ngx_wasm_preopen_t));
                if (wm->preopen == NULL) {
                    return NGX_CONF_ERROR;
                }
            }

            p = (u_char *) ngx_strchr(value[i].data, ':');
            if (p == NULL) {
                return "missing colon in preopen specification";
            }

            po = ngx_array_push(wm->preopen);
            if (po == NULL) {
                return NGX_CONF_ERROR;
            }

            po->host.data = value[i].data;
            po->host.len = p - value[i].data;

            if (po->host.len == 0) {
                return "missing host part in preopen specification";
            }

            po->host.data[po->host.len] = 0;

            if (po->host.len == value[i].len - 1) {
                return "missing guest part in preopen specification";
            }

            po->guest.data = p + 1;
            po->guest.len = value[i].data + value[i].len - po->guest.data;

            po->guest.data[po->guest.len] = 0;

            continue;
        }

        if (ngx_strncmp(value[i].data, "subsys=", 7) == 0) {
            value[i].data += 7;
            value[i].len -= 7;

            if (value[i].len == 0) {
                return "empty subsys value";
            }

#if defined (NGX_HTTP_WASM)
            if (value[i].len == 4
                && ngx_strncmp(value[i].data, "http", 4) == 0)
            {
                k = ngx_array_push(&wcf->http_modules);
                *k = wm->index;

                continue;
            }
#endif
            ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "unknown subsys: \"%V\"", &value[i]);

            return NGX_CONF_ERROR;
        }

        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "unknown parameter: \"%V\"", &value[i]);

        return NGX_CONF_ERROR;
    }

    if (wm->ns.len == 0) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0, "missing module id");
        return NGX_CONF_ERROR;
    }

    ngx_memzero(&file, sizeof(ngx_file_t));
    file.name = value[1];
    file.log = cf->log;

    file.fd = ngx_open_file(file.name.data, NGX_FILE_RDONLY, NGX_FILE_OPEN, 0);

    if (file.fd == NGX_INVALID_FILE) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, ngx_errno,
                           ngx_open_file_n " \"%V\" failed", &file.name);
        return NGX_CONF_ERROR;
    }

    if (ngx_fd_info(file.fd, &fi) == NGX_FILE_ERROR) {
        ngx_conf_log_error(NGX_LOG_CRIT, cf, ngx_errno,
                           ngx_fd_info_n " \"%V\" failed", &file.name);
        goto failed;
    }

    size = ngx_file_size(&fi);

    if (size == 0) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "\"%V\" zero file size", &file.name);
        goto failed;
    }

    buf = ngx_pnalloc(cf->pool, size);
    if (buf == NULL) {
        goto failed;
    }

    n = ngx_read_file(&file, buf, size, 0);

    if (n == NGX_ERROR) {
        ngx_conf_log_error(NGX_LOG_CRIT, cf, ngx_errno,
                           ngx_read_file_n " \"%V\" failed", &file.name);
        goto failed;
    }

    if ((size_t) n != size) {
        ngx_conf_log_error(NGX_LOG_CRIT, cf, 0,
                           ngx_read_file_n " \"%V\" returned only "
                           "%z bytes instead of %uz", &file.name, n, size);
        goto failed;
    }

    wm->src.data = buf;
    wm->src.len = n;

    if (ngx_close_file(file.fd) == NGX_FILE_ERROR) {
        ngx_log_error(NGX_LOG_ALERT, cf->log, ngx_errno,
                      ngx_close_file_n " \"%V\" failed", &file.name);
    }

    return NGX_CONF_OK;

failed:

    if (ngx_close_file(file.fd) == NGX_FILE_ERROR) {
        ngx_log_error(NGX_LOG_ALERT, cf->log, ngx_errno,
                      ngx_close_file_n " \"%V\" failed", &file.name);
    }

    return NGX_CONF_ERROR;
}


static void *
ngx_wasm_core_create_conf(ngx_cycle_t *cycle)
{
    ngx_wasm_conf_t  *wcf;

    wcf = ngx_palloc(cycle->pool, sizeof(ngx_wasm_conf_t));
    if (wcf == NULL) {
        return NULL;
    }

    wcf->use = NGX_CONF_UNSET_UINT;
    wcf->name = (void *) NGX_CONF_UNSET;

    if (ngx_array_init(&wcf->modules, cycle->pool, 4, sizeof(ngx_wasm_module_t))
        != NGX_OK)
    {
        return NGX_CONF_ERROR;
    }

#if defined(NGX_HTTP_WASM)

    if (ngx_array_init(&wcf->http_modules, cycle->pool, 4, sizeof(ngx_uint_t))
        != NGX_OK)
    {
        return NGX_CONF_ERROR;
    }

#endif

    return wcf;
}


static char *
ngx_wasm_core_init_conf(ngx_cycle_t *cycle, void *conf)
{
    ngx_wasm_conf_t  *wcf = conf;

    ngx_uint_t               i;
    ngx_module_t            *module;
    ngx_wasm_core_module_t  *wcm;

    module = NULL;

    for (i = 0; cycle->modules[i]; i++) {

        if (cycle->modules[i]->type != NGX_WASM_MODULE) {
            continue;
        }

        wcm = cycle->modules[i]->ctx;

        if (ngx_strcmp(wcm->name->data, ngx_wasm_core_module_name.data) == 0) {
            continue;
        }

        /*  stop search if we found configured module */
        if (cycle->modules[i]->ctx_index == wcf->use) {
            goto found;
        }

        /* remember first suitable module as default */
        if (module != NULL) {
            continue;
        }

        module = cycle->modules[i];
    }

    if (module == NULL) {
        ngx_log_error(NGX_LOG_EMERG, cycle->log, 0, "no wasm runtime found");
        return NGX_CONF_ERROR;
    }

    /* init default module */
    ngx_conf_init_uint_value(wcf->use, module->ctx_index);

    wcm = module->ctx;
    ngx_conf_init_ptr_value(wcf->name, wcm->name->data);

found:

    ngx_wasm_actions = wcm->actions;

    return NGX_CONF_OK;
}


static ngx_int_t
ngx_wasm_core_module_init(ngx_cycle_t *cycle)
{
    void             ***cf;
    ngx_wasm_conf_t    *wcf;

    cf = ngx_get_conf(cycle->conf_ctx, ngx_wasm_module);
    if (cf == NULL) {
        return NGX_OK;
    }

    wcf = (*cf)[ngx_wasm_core_module.ctx_index];

    if (!ngx_test_config && ngx_process <= NGX_PROCESS_MASTER) {
        ngx_log_error(NGX_LOG_NOTICE, cycle->log, 0,
                      "using the \"%s\" wasm runtime", wcf->name);
    }

    return NGX_OK;
}


static ngx_int_t
ngx_wasm_core_process_init(ngx_cycle_t *cycle)
{
    ngx_uint_t          i;
    ngx_wasm_conf_t    *wcf;
    ngx_wasm_module_t  *wm;

    wcf = ngx_wasm_get_conf(cycle->conf_ctx, ngx_wasm_core_module);

    if (wcf == NULL) {
        return NGX_OK;
    }

    if (wcf->modules.nelts == 0) {
        return NGX_OK;
    }

    if (ngx_wasm_create_runtime(cycle) != NGX_OK) {
        return NGX_ERROR;
    }

    wm = wcf->modules.elts;

    for (i = 0; i < wcf->modules.nelts; i++) {

        if (ngx_wasm_create_module(cycle, &wm[i]) != NGX_OK) {
            return NGX_ERROR;
        }

        if (wm[i].is_reactor) {
            wm[i].reactor = ngx_wasm_create_instance(&wm[i], cycle->pool,
                                                     cycle->log);
            if (wm[i].reactor == NULL) {
                return NGX_ERROR;
            }
        }
    }

    return NGX_OK;
}


static void
ngx_wasm_core_process_exit(ngx_cycle_t *cycle)
{
    ngx_uint_t          i;
    ngx_wasm_conf_t    *wcf;
    ngx_wasm_module_t  *wm;

    wcf = ngx_wasm_get_conf(cycle->conf_ctx, ngx_wasm_core_module);

    if (wcf == NULL) {
        return;
    }

    if (wcf->modules.nelts == 0) {
        return;
    }

    wm = wcf->modules.elts;

    for (i = 0; i < wcf->modules.nelts; i++) {
        if (wm[i].is_reactor && wm[i].reactor) {
            ngx_wasm_delete_instance(wm[i].reactor);
        }
        ngx_wasm_delete_module(&wm[i]);
    }

    ngx_wasm_delete_runtime(cycle);
}


char *
ngx_conf_set_wasm_ref_slot(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    char  *p = conf;

    ngx_str_t             *value;
    ngx_conf_post_t       *post;
    ngx_wasm_func_conf_t  *fc;

    fc = (ngx_wasm_func_conf_t *) (p + cmd->offset);

    if (fc->func.len) {
        return "is duplicate";
    }

    value = cf->args->elts;

    fc->wm = ngx_wasm_get_module(cf, &value[1]);
    if (fc->wm == NULL) {
        ngx_conf_log_error(NGX_LOG_EMERG, cf, 0,
                           "module \"%V\" not found", &value[1]);
        return NGX_CONF_ERROR;
    }

    fc->func.data = ngx_pnalloc(cf->pool, value[2].len + 1);
    if (fc->func.data == NULL) {
        return NGX_CONF_ERROR;
    }

    ngx_memcpy(fc->func.data, value[2].data, value[2].len);
    fc->func.len = value[2].len;

    /* zero-terminate function name for convenience */
    fc->func.data[fc->func.len] = 0;


    if (cmd->post) {
        post = cmd->post;
        return post->post_handler(cf, post, fc);
    }

    return NGX_CONF_OK;
}

