
/*
 * Copyright (C) 2024 Web Server LLC
 */

#include "ngx_wasm_serial.h"


static uint8_t *ngx_encode_array_head(ngx_arr_t *arr,
    ngx_arr_head_iter_t *iter, uint8_t *p, uint8_t *end);
static uint8_t *ngx_decode_array_head(ngx_arr_t *arr,
    ngx_arr_head_iter_t *iter, uint8_t *p, uint8_t *end);

static int32_t ngx_arr_alloc(ngx_arr_t *arr, ngx_alloc_ctx_t *ac);


static inline size_t
ngx_buf_advance(uint8_t *p, uint8_t *end, size_t len, size_t off)
{
    size_t  buf_size, left;

    buf_size = end - p;
    left = len - off;

    return buf_size < left ? buf_size : left;
}


uint8_t *
ngx_encode_len(uint64_t src, ngx_len_iter_t *iter, uint8_t *p, uint8_t *end)
{
    size_t   size, sh;
    uint8_t  i;

    switch (iter->state) {

    case ngx_len_st_init:

        if (end - p < 1) {
            return p;
        }

        iter->off = 0;

         if (src <= 63) {
            *p++ = src;
            iter->state = ngx_len_st_done;
            return p;
        }

        if (src <= 16383) {
            *p++ = 0x40 | (src >> 8);
            iter->len = 2;

        } else if (src <= 1073741823) {
            *p++ = 0x80 | (src >> 24);
            iter->len = 4;

        } else {
            *p++ = 0xc0 | (src >> 56);
            iter->len = 8;
        }

        iter->off = 1;

        iter->state = ngx_len_st_len;
        /* fall through */

    case ngx_len_st_len:

        size = ngx_buf_advance(p, end, iter->len, iter->off);

        if (size == 0) {
            return p;
        }

        for (i = 0; i < size; i++) {

            sh = (iter->len - iter->off - 1) * 8;

            *p++ = (src >> sh);

            iter->off++;
        }

        if (iter->off != iter->len) {
            return p;
        }

        iter->state = ngx_len_st_done;
        /* fall through */

    case ngx_len_st_done:
        return p;
    }

    return p;
}


uint8_t *
ngx_decode_len(uint64_t *dst, ngx_len_iter_t *iter, uint8_t *p, uint8_t *end)
{
    size_t   size;
    uint8_t  i;

    switch (iter->state) {

    case ngx_len_st_init:

        if (end - p < 1) {
            return p;
        }

        iter->off = 0;

        iter->len = 1 << (*p >> 6);
        *dst = *p++ & 0x3f;
        iter->off++;

        if (iter->off == iter->len) {
            iter->state = ngx_len_st_done;
            return p;
        }

        /* empty string { 0, NULL } */
        if (iter->len == 1 && (*dst) == 0) {
            iter->state = ngx_len_st_done;
            return p;
        }

        iter->state = ngx_len_st_len;
        /* fall through */

    case ngx_len_st_len:

        size = ngx_buf_advance(p, end, iter->len, iter->off);

        if (size == 0) {
            return p;
        }

        for (i = 0; i < size; i++) {
            *dst = ((*dst) << 8) + *p++;
        }

        iter->off += size;

        if (iter->off != iter->len) {
            return p;
        }

        iter->state = ngx_len_st_done;
        /* fall through */

    case ngx_len_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


uint8_t *
ngx_encode_str(ngx_str_t *src, ngx_str_iter_t *iter, uint8_t *p, uint8_t *end)
{
    size_t  size;

    if (src == NULL) {
        /* skip NULL string */
        iter->state = ngx_str_st_done;
        return p;
    }

    switch (iter->state) {

    case ngx_str_st_init:
        iter->off = 0;
        ngx_iter_reset(&iter->len_iter);

        iter->state = ngx_str_st_len;
        /* fall through */

    case ngx_str_st_len:

        p = ngx_encode_len(src->len, &iter->len_iter, p, end);

        if (!ngx_iter_done(&iter->len_iter)) {
            return p;
        }

        iter->state = ngx_str_st_data;
        /* fall through */

    case ngx_str_st_data:

        if (src->len == 0) {
            iter->state = ngx_str_st_done;
            /* empty string */
            return p;
        }

        size = ngx_buf_advance(p, end, src->len, iter->off);

        if (size == 0) {
            return p;
        }

        memcpy(p, src->data + iter->off, size);

        p += size;
        iter->off += size;

        if (iter->off != src->len) {
            return p;
        }

        iter->state = ngx_str_st_done;
        /* fall through */

    case ngx_str_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}



static uint8_t *
ngx_decode_str_real(ngx_str_t *dst, ngx_str_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac, uint8_t tzero)
{
    size_t  size;

    switch (iter->state) {

    case ngx_str_st_init:

        ngx_iter_reset(&iter->len_iter);
        iter->off = 0;
        iter->len = 0;

        iter->state = ngx_str_st_len;
        /* fall through */

    case ngx_str_st_len:

        p = ngx_decode_len(&iter->len, &iter->len_iter, p, end);
        if (!ngx_iter_done(&iter->len_iter)) {
            return p;
        }

        dst->len = iter->len;

        /* empty string */
        if (dst->len == 0) {
            dst->data = NULL;
            iter->state = ngx_str_st_done;
            return p;
        }

        if (ac->pool == NULL) {
            ac->pool = ngx_create_pool(dst->len + tzero, ac->log);
            if (ac->pool == NULL) {
                return NULL;
            }
        }

        dst->data = ngx_palloc(ac->pool, dst->len + tzero);
        if (dst->data == NULL) {
            return NULL;
        }

        iter->state = ngx_str_st_data;
        /* fall through */

    case ngx_str_st_data:

        size = ngx_buf_advance(p, end, dst->len, iter->off);
        if (size == 0) {
            return p;
        }

        memcpy(dst->data + iter->off, p, size);

        p += size;
        iter->off += size;

        if (iter->off != dst->len) {
            return p;
        }

        if (tzero) {
            dst->data[dst->len] = 0;
        }

        iter->state = ngx_str_st_done;
        /* fall through */

    case ngx_str_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


uint8_t *
ngx_decode_strz(ngx_str_t *dst, ngx_str_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac)
{
    return ngx_decode_str_real(dst, iter, p, end, ac, 1);
}


uint8_t *
ngx_decode_str(ngx_str_t *dst, ngx_str_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac)
{
    return ngx_decode_str_real(dst, iter, p, end, ac, 0);
}


uint8_t *
ngx_encode_kv(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p, uint8_t *end)
{
    switch (iter->state) {

    case ngx_kv_st_init:
        ngx_iter_reset(&iter->vi);

        iter->state = ngx_kv_st_key;
        /* fall through */

    case ngx_kv_st_key:

        p = ngx_encode_str(&kv->key, &iter->vi, p, end);
        if (!ngx_iter_done(&iter->vi)) {
            return p;
        }

        ngx_iter_reset(&iter->vi);

        iter->state = ngx_kv_st_value;
        /* fall through */

    case ngx_kv_st_value:

        p = ngx_encode_str(&kv->value, &iter->vi, p, end);
        if (!ngx_iter_done(&iter->vi)) {
            return p;
        }

        iter->state = ngx_kv_st_done;
        /* fall through */

    case ngx_kv_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


static uint8_t *
ngx_decode_kv_real(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac, uint8_t zero)
{
    switch (iter->state) {

    case ngx_kv_st_init:

        ngx_iter_reset(&iter->vi);

        iter->state = ngx_kv_st_key;
        /* fall through */

    case ngx_kv_st_key:

        p = zero ? ngx_decode_strz(&kv->key, &iter->vi, p, end, ac)
                 : ngx_decode_str(&kv->key, &iter->vi, p, end, ac);

        if (p == NULL) {
            return NULL;
        }

        if (iter->vi.state != ngx_str_st_done) {
            return p;
        }

        ngx_iter_reset(&iter->vi);

        iter->state = ngx_kv_st_value;
        /* fall through */

    case ngx_kv_st_value:

        p = zero ? ngx_decode_strz(&kv->value, &iter->vi, p, end, ac)
                 : ngx_decode_str(&kv->value, &iter->vi, p, end, ac);
        if (p == NULL) {
            return NULL;
        }

        if (iter->vi.state != ngx_str_st_done) {
            return p;
        }

        iter->state = ngx_kv_st_done;
        /* fall through */

    case ngx_kv_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


uint8_t *
ngx_decode_kv(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac)
{
    return ngx_decode_kv_real(kv, iter, p, end, ac, 0);
}


uint8_t *
ngx_decode_kvz(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac)
{
    return ngx_decode_kv_real(kv, iter, p, end, ac, 1);
}


static uint8_t *
ngx_encode_array_head(ngx_arr_t *arr, ngx_arr_head_iter_t *hi, uint8_t *p,
    uint8_t *end)
{
    switch (hi->state) {

    case ngx_arr_head_st_init:
        ngx_iter_reset(&hi->len_iter);

        hi->state = ngx_arr_head_st_nelts;
        /* fall through */

    case ngx_arr_head_st_nelts:

        p = ngx_encode_len(arr->nelts, &hi->len_iter, p, end);

        if (!ngx_iter_done(&hi->len_iter)) {
            return p;
        }

        ngx_iter_reset(&hi->len_iter);

        hi->state = ngx_arr_head_st_size;
        /* fall through */

    case ngx_arr_head_st_size:

        p = ngx_encode_len(arr->size, &hi->len_iter, p, end);

        if (!ngx_iter_done(&hi->len_iter)) {
            return p;
        }

        hi->state = ngx_arr_head_st_done;
        /* fall through */

    case ngx_arr_head_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


static uint8_t *
ngx_decode_array_head(ngx_arr_t *arr, ngx_arr_head_iter_t *hi, uint8_t *p,
    uint8_t *end)
{
    switch (hi->state) {
    case ngx_arr_head_st_init:
        ngx_iter_reset(&hi->len_iter);

        hi->state = ngx_arr_head_st_nelts;
        /* fall through */

    case ngx_arr_head_st_nelts:

        p = ngx_decode_len(&arr->nelts, &hi->len_iter, p, end);
        if (!ngx_iter_done(&hi->len_iter)) {
            return p;
        }

        ngx_iter_reset(&hi->len_iter);

        hi->state = ngx_arr_head_st_size;
        /* fall through */

    case ngx_arr_head_st_size:

        p = ngx_decode_len(&arr->size, &hi->len_iter, p, end);

        if (!ngx_iter_done(&hi->len_iter)) {
            return p;
        }

        hi->state = ngx_arr_head_st_done;
        /* fall through */

    case ngx_arr_head_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


uint8_t *
ngx_encode_array(ngx_arr_t *arr, ngx_arr_iter_t *ai, uint8_t *p, uint8_t *end)
{
    int32_t     rc;
    ngx_elt_t  *elt;

    switch (ai->state) {
    case ngx_array_st_init:

        ngx_iter_reset(&ai->head_iter);

        ai->off = 0;

        ai->state = ngx_array_st_head;
        /* fall through */

    case ngx_array_st_head:

        p = ngx_encode_array_head(arr, &ai->head_iter, p, end);

        if (!ngx_iter_done(&ai->head_iter)) {
            return p;
        }

        ai->state = ngx_array_st_items;
        ngx_iter_reset((ngx_any_iter_t *) arr->elt.iter);

        /* fall through */

    case ngx_array_st_items:

        elt = &arr->elt;

        while (ai->off < arr->nelts) {

            rc = arr->curr(ai->off, arr->data, elt->tmp);
            if (rc < 0) {
                return NULL;
            }

            p = elt->encode(elt->tmp, elt->iter, p, end);
            if (!ngx_iter_done((ngx_any_iter_t *) elt->iter)) {
                return p;
            }

            ngx_iter_reset((ngx_any_iter_t *) elt->iter);
            ai->off++;

            if (arr->next) {

                if (ai->off == arr->nelts) {
                    break;
                }

                rc = arr->next(arr->data);
                if (rc < 0) {
                    return NULL;
                }
            }

        }

        ai->state = ngx_array_st_done;

        /* fall through */

    case ngx_array_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


uint8_t *
ngx_decode_array(ngx_arr_t *arr, ngx_arr_iter_t *ai, uint8_t *p, uint8_t *end,
    ngx_alloc_ctx_t *ac)
{
    int32_t     rc;
    ngx_elt_t  *elt;

    switch (ai->state) {

    case ngx_array_st_init:
        ngx_iter_reset(&ai->head_iter);
        ai->off = 0;

        ai->state = ngx_array_st_head;
        /* fall through */

    case ngx_array_st_head:
        p = ngx_decode_array_head(arr, &ai->head_iter, p, end);

        if (!ngx_iter_done(&ai->head_iter)) {
            return p;
        }

        if (arr->dsize && ngx_arr_alloc(arr, ac) < 0) {
            return NULL;
        }

        ai->state = ngx_array_st_items;

        ngx_iter_reset((ngx_any_iter_t *) arr->elt.iter);

        /* fall through */

    case ngx_array_st_items:

        elt = &arr->elt;

        while (ai->off < arr->nelts) {

            p = elt->decode(elt->tmp, elt->iter, p, end, ac);
            if (p == NULL) {
                return NULL;
            }

            if (!ngx_iter_done((ngx_any_iter_t *) elt->iter)) {
                return p;
            }

            rc = arr->init(arr, ai->off, elt->tmp);
            if (rc < 0) {
                return NULL;
            }

            ngx_iter_reset((ngx_any_iter_t *) elt->iter);
            ai->off++;
        }

        ai->state = ngx_array_st_done;

        /* fall through */

    case ngx_array_st_done:
        return p;

    default:
        return NULL;
    }

    return p;
}


static int32_t
ngx_arr_alloc(ngx_arr_t *arr, ngx_alloc_ctx_t *ac)
{
    size_t  arr_len, len;

    arr_len = arr->dsize * arr->nelts;

    len = arr_len + arr->size;

    if (arr->zero) {
        len += arr->nelts * arr->zero;
    }

    ac->pool = ngx_create_pool(len, ac->log);
    if (ac->pool == NULL) {
        return -1;
    }

    arr->elts = ngx_palloc(ac->pool, arr_len);
    if (arr->elts == NULL) {
        return -1;
    }

    return 0;
}


void
ngx_str_array_init(ngx_str_array_t *sa, uint8_t zero)
{
    ngx_memzero(sa, sizeof(ngx_str_array_t));

    sa->arr.elt.tmp = &sa->str;
    sa->arr.elt.iter = &sa->str_iter;
    sa->arr.elt.encode = (ngx_elt_enc_pt) ngx_encode_str;

    if (zero) {
        sa->arr.elt.decode = (ngx_elt_dec_pt) ngx_decode_strz;

    } else {
        sa->arr.elt.decode = (ngx_elt_dec_pt) ngx_decode_str;
    }
}


static int32_t
ngx_get_str(uint64_t off, void *data, void *elt)
{
    ngx_str_t *str;
    ngx_str_t *values;

    str = elt;
    values = data;

    *str = values[off];

    return 0;
}


void
ngx_str_array_init_array_encoder(ngx_str_array_t *sa, uint64_t nelts,
    ngx_str_t *items)
{
    uint64_t  i;

    ngx_str_array_init(sa, 0);

    for (i = 0; i < nelts; i++) {
        sa->arr.size += items[i].len;
    }

    sa->arr.nelts = nelts;
    sa->arr.data = items;
    sa->arr.curr = ngx_get_str;
}


void
ngx_kv_array_init(ngx_kv_array_t *ka, uint8_t zero)
{
    ngx_memzero(ka, sizeof(ngx_kv_array_t));

    ka->arr.elt.tmp = &ka->kv;
    ka->arr.elt.iter = &ka->kv_iter;
    ka->arr.elt.encode = (ngx_elt_enc_pt) ngx_encode_kv;
    if (zero) {
        ka->arr.elt.decode = (ngx_elt_dec_pt) ngx_decode_kvz;

    } else {
        ka->arr.elt.decode = (ngx_elt_dec_pt) ngx_decode_kv;
    }
}


static int32_t
ngx_get_kv(uint64_t off, void *data, void *elt)
{
    ngx_kv_t  *kvs, *kv;

    kv = elt;
    kvs = data;

    *kv = kvs[off];

    return 0;
}


void
ngx_kv_array_init_array_encoder(ngx_kv_array_t *ka, uint64_t nelts,
    ngx_kv_t *kvs)
{
    uint64_t  i;

    ngx_kv_array_init(ka, 0);

    for (i = 0; i < nelts; i++) {
        ka->arr.size += kvs[i].key.len + kvs[i].value.len;
    }

    ka->arr.nelts = nelts;
    ka->arr.data = kvs;
    ka->arr.curr = ngx_get_kv;
}
