
/*
 * Copyright (C) 2024 Web Server LLC
 */

#ifndef _NGX_WASM_SERIAL_H_INCLUDED_
#define _NGX_WASM_SERIAL_H_INCLUDED_

#include <ngx_config.h>
#include <ngx_core.h>


#define NGX_ITER_ST_INITIAL 0     /* memzero on iter is sufficient to reset */
#define NGX_ITER_ST_FINAL   0xDD  /* final state for all iterators to test */

#define ngx_iter_reset(it)  (it)->state = NGX_ITER_ST_INITIAL
#define ngx_iter_done(it)  ((it)->state == NGX_ITER_ST_FINAL)


typedef enum {
    ngx_len_st_init = NGX_ITER_ST_INITIAL,
    ngx_len_st_len,
    ngx_len_st_done = NGX_ITER_ST_FINAL
} ngx_len_state_t;

typedef enum {
    ngx_str_st_init = NGX_ITER_ST_INITIAL,
    ngx_str_st_len,
    ngx_str_st_data,
    ngx_str_st_done = NGX_ITER_ST_FINAL
} ngx_str_state_t;

typedef enum {
    ngx_kv_st_init = NGX_ITER_ST_INITIAL,
    ngx_kv_st_key,
    ngx_kv_st_value,
    ngx_kv_st_done = NGX_ITER_ST_FINAL
} ngx_kv_state_t;

typedef enum {
    ngx_arr_head_st_init = NGX_ITER_ST_INITIAL,
    ngx_arr_head_st_nelts,
    ngx_arr_head_st_size,
    ngx_arr_head_st_done = NGX_ITER_ST_FINAL
} ngx_arr_head_state_t;

typedef enum {
    ngx_array_st_init = NGX_ITER_ST_INITIAL,
    ngx_array_st_head,
    ngx_array_st_items,
    ngx_array_st_done = NGX_ITER_ST_FINAL
} ngx_arr_state_t;


typedef struct {
    uint8_t               state;
} ngx_any_iter_t;

typedef struct {
    ngx_len_state_t       state;
    uint8_t               len;       /* varint size: 1..8 bytes */
    uint8_t               off;       /* current offset in size  */
} ngx_len_iter_t;

typedef struct {
    ngx_str_state_t       state;
    ngx_len_iter_t        len_iter;  /* str->len state  */
    uint64_t              off;       /* str->data state */
    uint64_t              len;       /* decoded length */
} ngx_str_iter_t;

typedef struct {
    ngx_kv_state_t        state;
    ngx_str_iter_t        vi;        /* key, then value state */
} ngx_kv_iter_t;

typedef struct {
    ngx_arr_head_state_t  state;
    ngx_len_iter_t        len_iter;  /* nelts, size */
} ngx_arr_head_iter_t;

typedef struct {
    ngx_arr_state_t       state;
    ngx_arr_head_iter_t   head_iter; /* array header */
    uint64_t              off;       /* curr array index */
} ngx_arr_iter_t;


typedef struct {
    ngx_str_t             key;
    ngx_str_t             value;
} ngx_kv_t;

typedef struct ngx_arr_s ngx_arr_t;

typedef struct {
    uint8_t              *reuse;     /* previously allocated, for realloc */
    ngx_pool_t           *pool;
    ngx_log_t            *log;
} ngx_alloc_ctx_t;


typedef uint8_t *(*ngx_elt_enc_pt)(void *item, void *iter,
    uint8_t *p, uint8_t *end);

typedef uint8_t *(*ngx_elt_dec_pt)(void *item, void *iter,
    uint8_t *p, uint8_t *end, ngx_alloc_ctx_t *ac);

typedef int32_t (*ngx_elt_curr_pt)(uint64_t i, void *ctx, void *elt);
typedef int32_t (*ngx_elt_next_pt)(void *data);

typedef int32_t (*ngx_elt_init_pt)(ngx_arr_t *arr, uint64_t off, void *src);


typedef struct ngx_elt_s {
    void                 *iter;        /* pointer to used iterator */
    void                 *tmp;         /* pointer to element */
    ngx_elt_dec_pt        decode;      /* decoder method */
    ngx_elt_enc_pt        encode;      /* encoder method */
} ngx_elt_t;


struct ngx_arr_s {
    uint64_t              nelts;       /* number of elements in array */
    uint64_t              size;        /* mem size needed to save data */

    ngx_elt_t             elt;         /* element spec */

    /* encoder */
    ngx_elt_curr_pt       curr;        /* get current item */
    ngx_elt_next_pt       next;        /* select next item */

    /* decoder */
    void                 *elts;        /* allocated array */
    ngx_elt_init_pt       init;        /* set dst entry from decoded element */
    size_t                dsize;
    uint8_t               zero;        /* null-terminate decoded strings */

    void                 *data;        /* opaque data for interface user */
};


/* commonly used types */

typedef struct {
    ngx_arr_t             arr;
    ngx_arr_iter_t        ai;
    ngx_str_t             str;
    ngx_str_iter_t        str_iter;
    ngx_alloc_ctx_t       ac;
} ngx_str_array_t;


typedef struct {
    ngx_arr_t             arr;
    ngx_arr_iter_t        ai;

    ngx_kv_t              kv;
    ngx_kv_iter_t         kv_iter;

    ngx_alloc_ctx_t       ac;
} ngx_kv_array_t;


uint8_t *ngx_encode_len(uint64_t src, ngx_len_iter_t *iter, uint8_t *p,
    uint8_t *end);
uint8_t *ngx_decode_len(uint64_t *dst, ngx_len_iter_t *iter, uint8_t *p,
    uint8_t *end);

uint8_t *ngx_encode_str(ngx_str_t *src, ngx_str_iter_t *iter, uint8_t *p,
    uint8_t *end);
uint8_t *ngx_decode_str(ngx_str_t *dst, ngx_str_iter_t *iter, uint8_t *p,
    uint8_t *end, ngx_alloc_ctx_t *ac);
uint8_t *ngx_decode_strz(ngx_str_t *dst, ngx_str_iter_t *iter, uint8_t *p,
    uint8_t *end, ngx_alloc_ctx_t *ac);

uint8_t *ngx_encode_kv(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p,
    uint8_t *end);
uint8_t *ngx_decode_kv(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p,
    uint8_t *end, ngx_alloc_ctx_t *ac);
uint8_t *ngx_decode_kvz(ngx_kv_t *kv, ngx_kv_iter_t *iter, uint8_t *p,
    uint8_t *end, ngx_alloc_ctx_t *ac);

uint8_t *ngx_encode_array(ngx_arr_t *arr, ngx_arr_iter_t *iter,
    uint8_t *p, uint8_t *end);
uint8_t *ngx_decode_array(ngx_arr_t *arr, ngx_arr_iter_t *ai, uint8_t *p,
    uint8_t *end, ngx_alloc_ctx_t *ac);

void ngx_str_array_init(ngx_str_array_t *sa, uint8_t zero);
void ngx_kv_array_init(ngx_kv_array_t *ka, uint8_t zero);

void ngx_str_array_init_array_encoder(ngx_str_array_t *sa, uint64_t nelts,
    ngx_str_t *items);
void ngx_kv_array_init_array_encoder(ngx_kv_array_t *ka, uint64_t nelts,
    ngx_kv_t *kvs);

#endif /* _NGX_WASM_SERIAL_H_INCLUDED_ */
